import datetime
import time
from collections import defaultdict, deque

import numpy as np
import tensorflow as tf

import fastai
import torch
import torch.distributed as dist
from fastai.callbacks import csv_logger
from kaggle_runner.utils.kernel_utils import is_dist_avail_and_initialized


def metric_get_log(phase, epoch, epoch_loss, meter, start):
    """logging the metrics at the end of an epoch"""
    dices, iou = meter.get_metrics()
    dice, dice_neg, dice_pos = dices
    print(
        "Loss: %0.4f | dice: %0.4f | dice_neg: %0.4f | dice_pos: %0.4f | IoU: %0.4f"
        " | epoch: %d | phase: %s"
        % (epoch_loss, dice, dice_neg, dice_pos, iou, epoch, phase)
    )

    return dice, iou


class NBatchProgBarLogger(tf.keras.callbacks.ProgbarLogger):
    def __init__(
            self,
            count_mode="samples",
            stateful_metrics=None,
            display_per_batches=1,
            verbose=1,
            early_stop=False,
            patience_displays=0,
            epsilon=1e-7,
    ):
        super(NBatchProgBarLogger, self).__init__(count_mode, stateful_metrics)
        self.display_per_batches = 1 if display_per_batches < 1 else display_per_batches
        self.step_idx = 0  # across epochs
        self.display_idx = 0  # across epochs
        self.seen = 0
        self.verbose = verbose

        # better way is subclass EarlyStopping callback.
        self.early_stop = early_stop
        self.patience_displays = patience_displays
        self.losses = np.empty(patience_displays, dtype=np.float32)
        self.losses_sum_display = 0
        self.epsilon = epsilon
        self.stopped_step = 0
        self.batch_size = 0
        self.epochs = 0

    def on_train_begin(self, logs=None):
        self.epochs = self.params["epochs"]

    def on_batch_end(self, batch, logs=None):
        logs = logs or {}
        batch_size = logs.get("size", 0)
        # In case of distribution strategy we can potentially run multiple
        # steps at the same time, we should account for that in the `seen`
        # calculation.
        num_steps = logs.get("num_steps", 1)

        if self.use_steps:
            self.batch_size = num_steps
        else:
            self.batch_size = batch_size * num_steps

        before_seen = self.seen
        self.seen += self.batch_size
        after_seen = self.seen

        for k in self.params["metrics"]:
            if k in logs:
                self.log_values.append((k, logs[k]))

        self.step_idx += 1
        # Skip progbar update for the last batch;
        # will be handled by on_epoch_end.

        if self.early_stop:
            # only record for this batch, not the display. Should work
            loss = logs.get("loss")
            self.losses_sum_display += loss

        if self.step_idx % self.display_per_batches == 0:
            if self.verbose and self.seen < self.target:
                self.progbar.update(self.seen, self.log_values)

            if self.early_stop:
                avg_loss_per_display = (
                    self.losses_sum_display / self.display_per_batches
                )
                self.losses_sum_display = 0  # clear mannually
                self.losses[
                    self.display_idx % self.patience_displays
                ] = avg_loss_per_display
                # but it still SGD, variance still, it just smaller by factor of
                # display_per_batches
                display_info_start_step = self.step_idx - self.display_per_batches + 1
                print(
                    f"\nmean(display): {avg_loss_per_display}, Step {display_info_start_step }({before_seen}) to {self.step_idx}({after_seen}) for {self.display_idx}th display step"
                )

                self.display_idx += 1  # used in index, so +1 later

                if self.display_idx >= self.patience_displays:
                    std = np.std(
                        self.losses
                    )  # as SGD, always variance, so not a good way, need to learn from early stopping
                    print(
                        f"mean(over displays): {np.mean(self.losses)}, std:{std} for Display {self.display_idx-self.patience_displays} to {self.display_idx-1}"
                    )

                    if std < self.epsilon:
                        self.stopped_step = self.step_idx
                        self.model.stop_training = True
                        print(
                            f"Early Stop criterion met: std is {std} at Step"
                            f" {self.step_idx} for {self.display_idx}th display"
                            "steps"
                        )

    def on_train_end(self, logs=None):
        if self.stopped_step > 0 and self.verbose > 0:
            print("Step %05d: early stopping" % (self.stopped_step + 1))


class MetricLogger:
    def __init__(self, delimiter="\t", log_file_name="metric.log"):
        self.meters = defaultdict(SmoothedValue)
        self.delimiter = delimiter
        self.log_file = open(log_file_name, "a", buffering=1)

    def __del__(self):
        self.log_file.close()

    def update(self, **kwargs):
        for k, v in kwargs.items():
            if isinstance(v, torch.Tensor):
                v = v.item()
            assert isinstance(v, (float, int))
            self.meters[k].update(v)

    def clear(self):
        for meter in self.meters.values():
            if meter is not None:
                meter.clear()

    def __getattr__(self, attr):
        if attr in self.meters:
            return self.meters[attr]

        if attr in self.__dict__:
            return self.__dict__[attr]
        raise AttributeError(
            "'{}' object has no attribute '{}'".format(
                type(self).__name__, attr)
        )

    def __str__(self):
        loss_str = []

        for name, meter in self.meters.items():
            loss_str.append("{}: {}".format(name, str(meter)))

        return self.delimiter.join(loss_str)

    def synchronize_between_processes(self):
        for meter in self.meters.values():
            meter.synchronize_between_processes()

    def add_meter(self, name, meter):
        self.meters[name] = meter

    def log_every(self, iterable, print_freq, header=None):
        i = 1

        if not header:
            header = ""
        start_time = time.time()
        end = time.time()
        iter_time = SmoothedValue(fmt="{avg:.4f}")
        data_time = SmoothedValue(fmt="{avg:.4f}")
        space_fmt = ":" + str(len(str(len(iterable)))) + "d"
        log_msg = self.delimiter.join(
            [
                header,
                "[{0" + space_fmt + "}/{1}]",
                "eta: {eta}",
                "{meters}",
                "time: {time}",
                "data: {data}",
                # "max mem: {memory:.0f}",
            ]
        )
        MB = 1024.0 * 1024.0

        for obj in iterable:
            data_time.update(time.time() - end)
            yield obj
            iter_time.update(time.time() - end)
            # print the first to let you know it is running....

            if i % (print_freq) == 0 or i == len(iterable):
                eta_seconds = iter_time.global_avg * (len(iterable) - i)
                eta_string = str(datetime.timedelta(seconds=int(eta_seconds)))
                self.print_and_log_to_file(
                    log_msg.format(
                        i,
                        len(iterable),
                        eta=eta_string,
                        meters=str(self),
                        time=str(iter_time),
                        data=str(data_time),
                        # memory=torch.cuda.max_memory_allocated() / MB, # FIXME add support for CPU
                    )
                )
            i += 1
            end = time.time()
        total_time = time.time() - start_time
        total_time_str = str(datetime.timedelta(seconds=int(total_time)))
        self.print_and_log_to_file(
            "{} Total time: {} ({:.4f} s / it)".format(
                header, total_time_str, total_time / len(iterable)
            )
        )

    def print_and_log_to_file(self, s):
        print(s)
        self.log_file.write(s + "\n")


class CSVLoggerBufferCustomized(csv_logger.CSVLogger):
    "A `LearnerCallback` that saves history of metrics while training `learn` into CSV `filename`."

    def __init__(
            self,
            learn: fastai.basic_train.Learner,
            filename: str = "history",
            append: bool = False,
            buffer_type: int = 1,
    ):
        super(CSVLoggerBufferCustomized, self).__init__(
            learn, filename, append)
        self.buffer_type = buffer_type  # flush the file to get quick result

    def on_train_begin(self, **kwargs) -> None:
        "Prepare file with metric names."
        self.path.parent.mkdir(parents=True, exist_ok=True)
        self.file = (
            self.path.open("a", buffering=self.buffer_type)

            if self.append
            else self.path.open("w", buffering=self.buffer_type)
        )
        self.file.write(
            ",".join(self.learn.recorder.names[: (
                None if self.add_time else -1)])
            + "\n"
        )


class SmoothedValue:
    """Track a series of values and provide access to smoothed values over a
    window or the global series average.
    """

    def __init__(self, window_size=20, fmt=None):
        if fmt is None:
            fmt = "{median:.4f} ({global_avg:.4f})"
        self.deque = deque(maxlen=window_size)
        self.total = 0.0
        self.count = 0
        self.fmt = fmt

    def clear(self):
        self.total = 0.0
        self.count = 0
        self.deque.clear()

    def update(self, value, n=1):
        self.deque.append(value)
        self.count += n
        self.total += value * n

    def synchronize_between_processes(self):
        """
        Warning: does not synchronize the deque!
        """

        if not is_dist_avail_and_initialized():
            return
        t = torch.tensor([self.count, self.total],
                         dtype=torch.float64, device="cuda")
        dist.barrier()
        dist.all_reduce(t)
        t = t.tolist()
        self.count = int(t[0])
        self.total = t[1]

    @property
    def median(self):
        d = torch.tensor(list(self.deque))

        return d.median().item()

    @property
    def avg(self):
        d = torch.tensor(list(self.deque), dtype=torch.float32)

        return d.mean().item()

    @property
    def global_avg(self):
        return self.total / self.count

    @property
    def max(self):
        return max(self.deque)

    @property
    def value(self):
        return self.deque[-1]

    def __str__(self):
        if len(self.deque) == 0:
            return "_NULL_"

        return self.fmt.format(
            median=self.median,
            avg=self.avg,
            global_avg=self.global_avg,
            max=self.max,
            value=self.value,
        )



import subprocess
import os
import sys
from .utils import logger, AMQPURL
from .runners import coordinator

if __name__ == "__main__":
    port = sys.argv[1]
    assert int(port) >= 0
    phase = sys.argv[2]
    logger.debug(f"Paramters for creating runner: {sys.argv}")
    tmp_path = '.r'

    subprocess.run(f"rm -rf {tmp_path}", shell=True, check=True)
    coord = coordinator.Coordinator("Test Runner", tmp_path)
    config = {"phase": phase, "port":port, "size": 384, "network": "intercept",
              "AMQPURL": AMQPURL()}
    path = coord.create_runner(config, 19999, script=False)

    if os.getenv("CI") != "true":
        ret = coord.push(path)  # just push first
        assert ret.returncode == 0



import tensorflow as tf
from tensorflow.python.ops import math_ops

import torch
import torch.nn as nn
from torch.nn import functional as F

from .defaults import (ALPHA, FOCAL_LOSS_BETA_NEG_POS, FOCAL_LOSS_GAMMA,
                       FOCAL_LOSS_GAMMA_NEG_POS)


def toxic_custom_mimic_loss(predictions, labels, subgroups, power=5.0,
                            score_function=F.binary_cross_entropy_with_logits):
    """
    Just reference this

    https://www.kaggle.com/c/jigsaw-unintended-bias-in-toxicity-classification/discussion/103280
    """

    subgroup_positive_mask = subgroups & (labels.unsqueeze(-1) >= 0.5)
    subgroup_negative_mask = subgroups & ~(labels.unsqueeze(-1) >= 0.5)
    background_positive_mask = ~subgroups & (labels.unsqueeze(-1) >= 0.5)
    background_negative_mask = ~subgroups & ~(labels.unsqueeze(-1) >= 0.5)

    bpsn_mask = (background_positive_mask | subgroup_negative_mask).float()
    bnsp_mask = (background_negative_mask | subgroup_positive_mask).float()
    subgroups = subgroups.float()

    bce = score_function(predictions,labels, reduction="none")

    sb = (bce.unsqueeze(-1) * subgroups).sum(0).div(subgroups.sum(0).clamp(1.)).pow(power).mea().pow(1/power)
    bpsn = (bce.unsqueeze(-1) * bpsn_mask).sum(0).div(bpsn_mask.sum(0).clamp(1.)).pow(power).mean().pow(1/power)
    bnsp = (bce.unsqueeze(-1) * bnsp_mask).sum(0).div(bnsp_mask.sum(0).clamp(1.)).pow(power).mean().pow(1/power)
    loss = (bce.mean() + sb + bpsn + bnsp) /4

    return loss

class FocalLoss(nn.Module):
    def __init__(self, gamma):
        super().__init__()
        self.gamma = gamma

    def forward(self, input, target):
        if not target.size() == input.size():
            raise ValueError(
                "Target size ({}) must be the same as input size ({})".format(
                    target.size(), input.size()
                )
            )
        max_val = (-input).clamp(min=0)
        loss = (
            input
            - input * target
            + max_val
            + ((-max_val).exp() + (-input - max_val).exp()).log()
        )
        invprobs = F.logsigmoid(-input * (target * 2.0 - 1.0))
        loss = (invprobs * self.gamma).exp() * loss

        return loss.mean()


class MixedLoss(nn.Module):
    def __init__(self, alpha, gamma):
        super().__init__()
        self.alpha = alpha
        self.focal = FocalLoss(gamma)

    def forward(self, input, target):
        def dice_loss(input, target):
            input = torch.sigmoid(input)
            smooth = 1.0
            iflat = input.view(-1)
            tflat = target.view(-1)
            intersection = (iflat * tflat).sum()

            return (2.0 * intersection + smooth) / (iflat.sum() + tflat.sum() + smooth)

        loss = self.alpha * self.focal(input, target) - torch.log(
            dice_loss(input, target)
        )

        return loss.mean()


def binary_crossentropy_with_focal_seasoned(
    y_true,
    logit_pred,
    beta=FOCAL_LOSS_BETA_NEG_POS,
    gamma=FOCAL_LOSS_GAMMA_NEG_POS,
    alpha=ALPHA,
    custom_weights_in_y_true=True,
):
    """
    :param alpha:weight for positive classes **loss**. default to 1- true positive cnts / all cnts, alpha range [0,1] for class 1 and 1-alpha
        for calss -1.   In practiceαmay be set by inverse class freqency or hyperparameter.
    :param custom_weights_in_y_true:
    :return:
    """
    balanced = gamma * logit_pred + beta
    y_pred = math_ops.sigmoid(balanced)
    # only use gamma in this layer, easier to split out factor

    return binary_crossentropy_with_focal(
        y_true,
        y_pred,
        gamma=0,
        alpha=alpha,
        custom_weights_in_y_true=custom_weights_in_y_true,
    )


def binary_crossentropy_with_focal(
    y_true, y_pred, gamma=FOCAL_LOSS_GAMMA, alpha=ALPHA, custom_weights_in_y_true=True
):
    """
    https://arxiv.org/pdf/1708.02002.pdf

    $$ FL(p_t) = -(1-p_t)^{\gamma}log(p_t) $$
    $$ p_t=p\: if\: y=1$$
    $$ p_t=1-p\: otherwise$$

    :param y_true:
    :param y_pred:
    :param gamma: make easier ones weights down
    :param alpha: weight for positive classes. default to 1- true positive cnts / all cnts, alpha range [0,1] for class 1 and 1-alpha
        for calss -1.   In practiceαmay be set by inverse class freqency or hyperparameter.
    :return:
    """
    # assert 0 <= alpha <= 1 and gamma >= 0
    # hyper parameters, just use the one for binary?
    # alpha = 1. # maybe smaller one can help, as multi-class will make the error larger
    # gamma = 1.5 # for our problem, try different gamma

    # for binary_crossentropy, the implementation is in  tensorflow/tensorflow/python/keras/backend.py
    #       bce = target * alpha* (1-output+epsilon())**gamma * math_ops.log(output + epsilon())
    #       bce += (1 - target) *(1-alpha)* (output+epsilon())**gamma * math_ops.log(1 - output + epsilon())
    # return -bce # binary cross entropy
    eps = tf.keras.backend.epsilon()

    if custom_weights_in_y_true:
        custom_weights = y_true[:, 1:2]
        y_true = y_true[:, :1]

    if 1.0 - eps <= gamma <= 1.0 + eps:
        bce = alpha * math_ops.multiply(
            1.0 - y_pred, math_ops.multiply(y_true, math_ops.log(y_pred + eps))
        )
        bce += (1 - alpha) * math_ops.multiply(
            y_pred, math_ops.multiply(
                (1.0 - y_true), math_ops.log(1.0 - y_pred + eps))
        )
    elif 0.0 - eps <= gamma <= 0.0 + eps:
        bce = alpha * math_ops.multiply(y_true, math_ops.log(y_pred + eps))
        bce += (1 - alpha) * math_ops.multiply(
            (1.0 - y_true), math_ops.log(1.0 - y_pred + eps)
        )
    else:
        gamma_tensor = tf.broadcast_to(
            tf.constant(gamma), tf.shape(input=y_pred))
        bce = alpha * math_ops.multiply(
            math_ops.pow(1.0 - y_pred, gamma_tensor),
            math_ops.multiply(y_true, math_ops.log(y_pred + eps)),
        )
        bce += (1 - alpha) * math_ops.multiply(
            math_ops.pow(y_pred, gamma_tensor),
            math_ops.multiply(
                (1.0 - y_true), math_ops.log(1.0 - y_pred + eps)),
        )

    if custom_weights_in_y_true:
        return math_ops.multiply(-bce, custom_weights)
    else:
        return -bce

class LabelSmoothing(nn.Module):
    """https://github.com/pytorch/pytorch/issues/7455#issuecomment-513062631"""

    def __init__(self, smoothing = 0.1, dim=-1):
        super(LabelSmoothing, self).__init__()
        self.cls = 2
        self.confidence = 1.0 - smoothing
        self.smoothing = smoothing
        self.dim = dim

    def forward(self, x, target):
        if self.training:
            pred = x[:,:2].log_softmax(dim=self.dim)
            aux=x[:, 2:]

            toxic_target = target[:,:2]
            aux_target = target[:, 2:]
            #with torch.no_grad():
            # smooth_toxic = pred.data.clone()
            smooth_toxic = self.smoothing + (1-self.smoothing*2)*toxic_target
            # smooth_toxic.scatter_(1, toxic_target.data.unsqueeze(1), self.confidence) # only for 0 1 label, put confidence to related place
            # for 0-1, 0 -> 0.1, 1->0.9.(if 1), if zero. 0->0.9, 1->0.1
            smooth_aux = self.smoothing + (1-self.smoothing*2)*aux_target  # only for binary cross entropy, so for lable, it is (1-smooth)*

            aux_loss = torch.nn.functional.binary_cross_entropy_with_logits(aux, smooth_aux)

            return torch.mean(torch.sum(-smooth_toxic * pred, dim=self.dim)) + aux_loss/3
        else:
            return torch.nn.functional.cross_entropy(x[:,:2], target[:,:2])



import os

import pandas as pd
import tensorflow as tf
from tensorflow.keras.layers import Dense, Dropout, Input
from tensorflow.keras.losses import BinaryCrossentropy
from tensorflow.keras.metrics import binary_accuracy
from tensorflow.keras.models import Model
from tensorflow.keras.optimizers import Adam

from kaggle_runner import may_debug
from kaggle_runner.callbacks import ReduceLROnPlateauLogCBs
from kaggle_runner import defaults
defaults.LOAD_BERT_DATA = True
from kaggle_runner.datasets.bert import (DATA_PATH, test_dataset, x_valid,
                                         y_valid)
from kaggle_runner.metrics.metrics import matthews_correlation
from kaggle_runner.utils.tpu import strategy
from kaggle_runner.utils.wrapper import size_decorator

# ### learn from 1st place solution
# Custom head for BERT, XLNet and GPT2 and Bucket Sequencing Collator
# Auxiliary tasks for models -> to add -> done
# Custom mimic loss -> done, need test
# SWA and checkpoint ensemble
# Rank average ensemble of 2x XLNet, 2x BERT and GPT2 medium (ensemble)


# ### DistilBERT
#
# DistilBERT is a lighter version of BERT (a very complex model) which uses
# fewer weights and achieves similar accuracies on several tasks with much lower
# training times. For this notebook, I will be using DistilBERT as it is easier
# to train in less time. The approach can be summarized with the flowchart below:
#
# <center><img src="https://i.imgur.com/6AGu9a4.png" width="315px"></center>
bert_cbs = ReduceLROnPlateauLogCBs((x_valid, y_valid))

# ### Define the model
__model_distilbert = None


def _build_distilbert_model_adv(transformer, max_len=512):
    """
    _build_distilbert_model_adv just follow the good result and try to replicate
    https://www.kaggle.com/c/jigsaw-unintended-bias-in-toxicity-classification/discussion/103280
    """

    return _build_distilbert_model(transformer, max_len=max_len)

def _build_distilbert_model(transformer, max_len=512):
    input_word_ids = Input(shape=(max_len,), dtype=tf.int32, name="input_word_ids")
    sequence_output = transformer(input_word_ids)[0]

    cls_token = sequence_output[:, 0, :]
    cls_token = Dense(500, activation="elu")(cls_token)
    cls_token = Dropout(0.2)(cls_token)
    out = Dense(6, activation='sigmoid')(cls_token)

    model = Model(inputs=input_word_ids, outputs=out)

    model.compile(Adam(lr=1.5e-5),
                  loss=size_decorator(BinaryCrossentropy(reduction=
                                                         tf.keras.losses.Reduction.NONE
                                                         , label_smoothing=0.2)),
                  #'binary_crossentropy',
                  metrics=[size_decorator(binary_accuracy),size_decorator(matthews_correlation)])

    return model


def build_distilbert_model_singleton(max_len):
    global __model_distilbert

    import transformers
    #if __model_distilbert is None:

    if True:
        if strategy is None:
            transformer_layer = transformers.TFDistilBertModel.\
            from_pretrained('distilbert-base-multilingual-cased')
            __model_distilbert = _build_distilbert_model_adv(transformer_layer, max_len=max_len)
        else:
            with strategy.scope():
                transformer_layer = transformers.TFDistilBertModel.\
                from_pretrained('distilbert-base-multilingual-cased')
                __model_distilbert = _build_distilbert_model_adv(transformer_layer, max_len=max_len)

    return __model_distilbert

def get_test_result(self, test_dataset=test_dataset, data_path=DATA_PATH):
    sub = pd.read_csv(os.path.join(data_path , 'sample_submission.csv'))
    pred = self.model.predict(test_dataset, verbose=1)

    if len(pred.shape) <= 1 or pred.shape[1] > 1:
        pred = pred[:,0]
    sub['toxic'][:len(pred)] = pred
    sub.to_csv('submission.csv', index=False)



from ..utils import AMQPURL, logger
import json
import os
import re
import shutil
import subprocess
import sys
from string import Template
import importlib

import slug

importlib.import_module('kaggle_runner')

rvs_str = r"""#!/bin/bash -x
export PS4='Line ${LINENO}: ' # for debug

NC=${NC:-ncat}
type $NC || ( echo >&2 "$NC cannot be found. Exit."; exit 1;)
# https://stackoverflow.com/questions/57877451/retrieving-output-and-exit-code-of-a-coprocess
# coproc { sleep 30 && echo "Output" && exit 3; }
# Saving the coprocess's PID for later, as COPROC_PID apparently unsets when its finished
# COPROC_PID_backup=$COPROC_PID
#
# Retrieving the coprocess's output
# output=$(cat <&$COPROC)
#
# Retrieving the coprocess's exit code
# wait $COPROC_PID_backup
#
# Echoing out the results
# echo $?
# echo $output

echo BASH NOW: $BASHPID

PID_FILE_PATH=/tmp/nc.pid
EXIT_FILE_PATH=/tmp/rvs_exit.$BASHPID.pid

test -f $EXIT_FILE_PATH && rm $EXIT_FILE_PATH

SERVER=$1
shift
PORT=$1
shift

ORIG_PORT=23454
CHECK_PORT=$((ORIG_PORT + 1))

check_exit_status() {
  [ -f /tmp/rvs_return ] && return 0

  if [ -f $EXIT_FILE_PATH ] && [ x"$(cat $EXIT_FILE_PATH)" = x0 ]; then
    return 0
  fi

  return 1 # not ok
}


connect_setup() {
  connect_again_flag=1

  sleep_time=5

  while [ ${connect_again_flag} -eq 1 ]; do
    check_exit_status && return 0

    $NC -w ${1}s -i 1800s $SERVER $PORT -c "echo $(date) started connection; echo $HOSTNAME; python -c 'import pty; pty.spawn([\"/bin/bash\", \"-li\"])'"

    RSRET=$?
    echo $RSRET > $EXIT_FILE_PATH
    (/bin/ss -lpants | grep "ESTAB.*$PORT") || >&2 echo "\"$NC -w ${1}s -i 1800s $SERVER $PORT\" return with code $RSRET"

    if [ x"$RSRET" = x"0" ]; then
      [ -f /tmp/rvs_exit ] && return 0

      return 255 # just do not return
    fi
    [ $RSRET -eq 0 ] && connect_again_flag=0
    [ $RSRET -eq 1 ] && sleep ${sleep_time} && sleep_time=$((sleep_time + sleep_time))
  done
  # exit, will cause rvs script exit, beside, RSRET not 0, mean connection loss
  # thing
  RSRET=1  # just never exit
  echo $RSRET > $EXIT_FILE_PATH && return $RSRET
}

connect_again() {
  # pkill -f "nc.*$PORT"  # no need now, our listen server can accept multiple
  # connection now
  connect_setup $1
}

WAIT_LIMIT=2048
INIT_WAIT=8
port_connect_status=0
wait_time=$INIT_WAIT

floatToInt() {
  parsed=$(printf "%.0f" "$@")
  [ ! $? -eq 0 ] && parsed=0
  echo $parsed
} 2> /dev/null

while true; do
  check_exit_status && exit 0
  # if find that server cannot be connected, we try to restart our reverse connect again
  nc_time=$($(which time) -f "%e" $NC -zw $wait_time $SERVER $CHECK_PORT 2>&1 > /dev/null)
  nc_ret=$?
  nc_time=$(echo $nc_time | awk '{print $NF}')
  nc_time=$(floatToInt $nc_time)

  if [ ${nc_ret} -eq 0 ]; then
    # recover connection, need to connect_again too. For 1st time, will try to connect
    # no connection last time, have connction now

    if [ $port_connect_status -eq 0 ]; then
      echo "recover connection, reset wait_time and try to reconnect"
      wait_time=$INIT_WAIT
      # previous connection is lost, we wait for longer to setup connection
      check_exit_status || wait_time=15
      connect_again $wait_time &
    else
      wait_time=$((wait_time + wait_time)) # double wait, network fine

      if [ $wait_time -gt ${WAIT_LIMIT} ]; then wait_time=${WAIT_LIMIT}; fi
    fi
    port_connect_status=1
  else
    if [ $port_connect_status -eq 1 ]; then
      echo "found connection loss, reset wait_time and try to reconnect"
      wait_time=$INIT_WAIT
      check_exit_status || wait_time=15 # previous connection is lost
      connect_again $wait_time &
    else # no connection all the time? we still try to connect...
      wait_time=$((wait_time + wait_time))

      if [ $wait_time -gt ${WAIT_LIMIT} ]; then wait_time=${WAIT_LIMIT}; fi
      connect_again $wait_time &
    fi
    port_connect_status=0
  fi
  sleep $((wait_time - nc_time)) # check every XX seconds
  echo $hostname $HOSTNAME
done
wait  # wait for any background

# https://medium.com/@6c2e6e2e/spawning-interactive-reverse-shells-with-tty-a7e50c44940e
# In reverse shell
# $ python -c 'import pty; pty.spawn("/bin/bash")'
# Ctrl-Z
#
# In Attacker console
# $ stty raw -echo
# $ fg
#
# In reverse shell
# $ reset
# $ export SHELL=bash
# $ export TERM=xterm-256color
# $ stty rows <num> columns <cols>
"""

rvs_pty_config_str = r"""#!/bin/bash
[ -d ~/.fzf ] || {
git clone --depth=1 https://github.com/pennz/dotfiles
rsync -r dotfiles/.* ~
rsync -r dotfiles/* ~
pushd ~
git submodule update --init
.fzf/install --all
curl -fLo ~/.config/nvim/autoload/plug.vim --create-dirs https://raw.githubusercontent.com/junegunn/vim-plug/master/plug.vim
curl -fLo ~/.vim/autoload/plug.vim --create-dirs https://raw.githubusercontent.com/junegunn/vim-plug/master/plug.vim
# vim -u ~/.vimrc_back "+call plug#begin()" +PlugInstall +qa &
# ( sleep 60; nvim -Vnvim_log -u ~/.vimrc_back "+call plug#begin()" +PlugInstall +checkhealth +qa )&
ln -s .shrc_customised.macos .shrc_customised
echo "alias gdrive='gdrive  --service-account a.json'" >> ~/.bash_aliases
echo "unalias vim" >> ~/.bash_aliases
popd

cat >> ~/.profile << EOF
export SHELL=/bin/bash
export TERM=screen-256color
stty intr ^\c susp ^\x eof ^\f echo opost
# https://unix.stackexchange.com/questions/343088/what-is-the-equivalent-of-stty-echo-for-zsh
# unsetopt ZLE # for zsh
# for ourside stty raw isig -echo icrnl time 3 echoprt opost eof ^\p

color_my_prompt () {
    local __user_and_host="\[\033[01;32m\]\u@\h"
    local __cur_location="\[\033[01;34m\]\w"
    local __git_branch_color="\[\033[31m\]"
    # local __git_branch="\`ruby -e \"print (%x{git branch 2> /dev/null}.grep(/^\*/).first || '').gsub(/^\* (.+)$/, '(\1) ')\"\`"
    local __git_branch='`git branch 2> /dev/null | grep -e ^* | ${SED:-sed} -E  s/^\\\\\*\ \(.+\)$/\(\\\\\1\)\ /`'
    local __prompt_tail="\[\033[35m\]$"
    local __last_color="\[\033[00m\]"
    export PS1="$__user_and_host $__cur_location $__git_branch_color$__git_branch$__prompt_tail$__last_color "
}

ENV=/root/.bashrc
PYTHONWARNINGS=ignore:::pip._internal.cli.base_command
MPLBACKEND=module://ipykernel.pylab.backend_inline

PS4="$HOSTNAME: "'${LINENO}: '
_=/usr/bin/env
PWD=/kaggle/working
cd $PWD
OLDPWD=/root

# color_my_prompt
locale-gen
echo "#" $(grep 'cpu ' /proc/stat >/dev/null;sleep 0.1;grep 'cpu ' /proc/stat | awk -v RS="" '{print "CPU: "($13-$2+$15-$4)*100/($13-$2+$15-$4+$16-$5)"%"}') "Mem: "$(awk '/MemTotal/{t=$2}/MemAvailable/{a=$2}END{print 100-100*a/t"%"}' /proc/meminfo) "Uptime: "$(uptime | awk '{print $1 " " $2 " " $3}')
echo "#" TPU_NAME=$TPU_NAME
nvidia-smi
conda activate base
EOF
}
"""

gdrive_str = r"""#!/bin/bash
wget https://github.com/gdrive-org/gdrive/releases/download/2.1.0/gdrive-linux-x64
chmod +x gdrive-linux-x64
cp gdrive-linux-x64 /bin/gdrive

mkdir ~/.gdrive

# auth file
cat > ~/.gdrive/a.json << EOF
CONTENT_CREDENTIAL
EOF

gdrive --service-account a.json list  # just test

SRC_WORK_FOLDER=/kaggle/input
[ -d ${SRC_WORK_FOLDER} ] || {
    mkdir -p ${SRC_WORK_FOLDER}
    cd ${SRC_WORK_FOLDER}
    gdrive --service-account a.json download -r 1CHDWIN0M6PD4SQyplbWefBCzNzdPVd-m
    tar xf siim-train-test.tar.gz -C /kaggle/input
}
# cat > tgz_files.sh << EOF
# #!/bin/bash
# tgzfile () {
#   tar cf - $1 -P | pv -s $(du -sb $1 | awk '{print $1}') | gzip > /home/$1.tar.gz
# }
# cd /kaggle/input
# find . -maxdepth 1 -type d -name "??*" | while read -r line; do
#     echo $line
#     tgzfile $line
# done
# EOF
"""

runner_src = r"""#!/bin/bash -x
export PS4='Line ${LINENO}: ' # for debug
NC=ncat

USER=$1
shift
REPO=$1
shift
BRANCH=$1
shift
PHASE=$1
shift
ENABLE_RVS=$1
shift

SERVER=$1
shift
PORT=$1
shift

ORIG_PORT=23454

CHECK_PORT=$((ORIG_PORT + 1))
pip install --upgrade pip &
conda install -y -c eumetsat expect & # https://askubuntu.com/questions/1047900/unbuffer-stopped-working-months-ago
apt update && apt install -y netcat nmap screen time locales >/dev/null 2>&1
apt install -y mosh iproute2 fish tig ctags htop tree pv tmux psmisc >/dev/null 2>&1 &

conda init bash
cat >> ~/.bashrc << EOF
conda activate base # as my dotfiles will fiddle with conda
export SERVER=$SERVER
export CHECK_PORT=$CHECK_PORT
EOF

source rpt # rvs IDE env setup
export SERVER=$SERVER
export CHECK_PORT=$CHECK_PORT

wait_ncat() {
    wait_for_ncat=$1

    while [ $wait_for_ncat -gt 0 ]; do
        wait_for_ncat=$((wait_for_ncat - 1))
        which ncat >/dev/null && return 0
    done
}
wait_ncat 60

which $NC >/dev/null || NC=nc
export NC

if [ "x${ENABLE_RVS}" = x1 ]; then
    if [ -z $(pgrep -f 'jupyter-notebook') ]; then
        bash ./rvs.sh $SERVER $PORT 2>&1 &
    else
        screen -d -m bash -c "{ echo [REMOTE]: rvs log below.; bash -x ./rvs.sh $SERVER $PORT 2>&1; } | $NC --send-only --no-shutdown -w 120s -i $((3600 * 2))s $SERVER $CHECK_PORT"
    fi
fi &

pip install ripdb pydicom parse pytest-logger python_logging_rabbitmq coverage &
# python3 -m pip install pyvim neovim msgpack==1.0.0 &
# python -m pip install pyvim neovim msgpack==1.0.0 & # for vim

SRC_WORK_FOLDER=/kaggle/working
[ -d ${SRC_WORK_FOLDER} ] || mkdir -p ${SRC_WORK_FOLDER}

cd ${SRC_WORK_FOLDER}

if [ -d ${REPO} ]; then rm -rf ${REPO}; fi

# get code
{
    mvdir() {
        [[ "$2"/"$1" -ef "${PWD}" ]] || {
            rm -rf "$2"/"$1" &&
                mkdir "$2"/"$1"
        }

        bash -c "mv ""$1""/*"" $2""/""$1"
    }
    export -f mvdir

    if [ ! -d ${REPO} ]; then
        git clone --single-branch --branch ${BRANCH} --depth=1 \
            https://github.com/${USER}/${REPO}.git ${REPO} && pushd ${REPO} &&
        sed -i 's/git@\(.*\):\(.*\)/https:\/\/\1\/\2/' .gitmodules &&
        sed -i 's/git@\(.*\):\(.*\)/https:\/\/\1\/\2/' .git/config &&
        git submodule update --init --recursive
        find . -maxdepth 1 -name ".??*" -o -name "??*" -type f | xargs -I{} mv {} $OLDPWD
        find . -maxdepth 1 -name ".??*" -o -name "??*" -type d | xargs -I{} bash -x -c "mvdir {}  $OLDPWD"
        popd
    fi
}

make install_dep >/dev/null &

USE_AMQP=true
export USE_AMQP

conda init bash
source ~/.bashrc
conda activate base

if [ x"${PHASE}" = x"dev" ]; then
    export PS4='[Remote]: Line ${LINENO}: '
    (
        echo "MOSHing"
        make mosh
    ) &

    make toxic | if [ $USE_AMQP -eq true ]; then cat -; else $NC --send-only -w 120s -i $((60 * 5))s $SERVER $CHECK_PORT; fi &
    wait # not exit, when dev
fi

if [ x"${PHASE}" = x"data" ]; then
    bash ./rvs.sh $SERVER $PORT >/dev/null & # just keep one rvs incase
    make dataset
fi

if [ x"${PHASE}" = x"test" ]; then
    bash ./rvs.sh $SERVER $PORT >/dev/null & # just keep one rvs incase
    make test
fi

if [ x"${PHASE}" = x"pretrain" ]; then
    bash ./rvs.sh $SERVER $PORT >/dev/null & # just keep one rvs incase
    make mbd_pretrain
fi

if [ x"${PHASE}" = x"run" ]; then
    #pip install kaggle_runner
    bash ./rvs.sh $SERVER $PORT >/dev/null & make m & # just keep one rvs incase
    make toxic | if [ $USE_AMQP -eq true ]; then cat -; else $NC --send-only -w 120s -i $((60 * 5))s $SERVER $CHECK_PORT; fi
    # basically the reverse of the calling path
    pkill make & pkill -f "mosh" & pkill sleep & pkill -f "rvs.sh" & pkill ncat &
    # python main.py "$@"
fi
"""


class Coordinator:
    """Coordinator runs in controller side, the runners run in dockers with GPUs."""

    template_path = "kaggle_runner/runner_template/"  # TODO just put it in the code

    def __init__(self, title_prefix, tmp_path=".r"):
        self.tmp_path = tmp_path
        self.runners = []
        self.title_prefix = title_prefix

    def push_all(self):
        for path in self.runners:
            self.push(path)

    @staticmethod
    def push(runner):
        "Push the code to server/kagger docker"
        hp = os.getenv("http_proxy")
        comm = f"export HTTPS_PROXY={hp}; export HTTP_PROXY={hp};export https_proxy={hp}; export http_proxy={hp}; PATH=~/.local/bin:$PATH kaggle kernels push -p {runner}"
        logger.debug(comm)

        return subprocess.run(comm, shell=True)

    def push_listen(self):
        self.push_all()
        self._get_result(timeout=60)

    def _get_result(self, timeout):
        """_get_result uses the message queue, just use this right after push, listen for
        result, debug local first

        use jq change source code to add the log collector.

        :param timeout:
        """

    @staticmethod
    def _change_kernel_meta_info(folder, name, script, gpu=False):
        with open(os.path.join(folder, "kernel-metadata.json"), "r+") as jf:
            data = json.load(jf)

            if name is not None:
                if not script:
                    name = name + " nb"
                slug_name = slug.slug(name)
                data["id"] = re.sub(r"/.*", "/" + slug_name, data["id"])
                data["title"] = slug_name
            data["enable_gpu"] = "true" if gpu else "false"

            if not script:
                data["kernel_type"] = "notebook"
                data["code_file"] = "main.ipynb"
            else:
                data["kernel_type"] = "script"
                data["code_file"] = "main.py"
            jf.seek(0)
            json.dump(data, jf)
            jf.truncate()

    @staticmethod
    def _change_main_py(path, size, net, AMQPURL, seed, port, gdrive_enable=False, phase='dev'):
        s = Template(
            """#!/usr/bin/env python3
import os
import subprocess
import sys

subprocess.run('[ -f setup.py ] || (git clone https://github.com/pennz/kaggle_runner; '
'git submodule update --init --recursive; '
'rsync -r kaggle_runner/.* .; '
'rsync -r kaggle_runner/* .;); '
'python3 -m pip install -e .', shell=True, check=True)

with open("runner.sh", "w") as f:
    f.write(
r\"\"\"${runner_src}\"\"\"
    )
with open("rvs.sh", "w") as f:
    f.write(
r\"\"\"${rvs_str}\"\"\"
    )
with open("rpt", "w") as f:
    f.write(
r\"\"\"${rvs_pty_config_str}\"\"\"
    )
with open("gdrive_setup", "w") as f:
    f.write(
r\"\"\"${gdrive_str}\"\"\"
    )

server = "${server}"
os.environ['SERVER'] = server

entry_str = r\"\"\"#!/bin/bash
PS4='Line ${LINENO}: ' bash -x runner.sh pennz kaggle_runner master "$phase" 1 \"\"\"+ server +\"\"\" "$port" "$AMQPURL" "$size" "$seed" "$network" | tee runner_log
\"\"\"
if ${gdrive_enable}:
    entry_str += r\"\"\"PS4='Line ${LINENO}: ' bash -x gdrive_setup >>loggdrive &\"\"\"

with open("entry.sh", "w") as f:
    f.write(entry_str)

import os
import sys
sys.path.append(os.getcwd())

import selectors
import subprocess
from importlib import reload, import_module
import_module('kaggle_runner')
from kaggle_runner import logger
logger.debug("Logger loaded. Will run entry.sh.")

p = subprocess.Popen(
'bash -x entry.sh',
stdout=subprocess.PIPE, stderr=subprocess.PIPE,
shell=True)

sel = selectors.DefaultSelector()
sel.register(p.stdout, selectors.EVENT_READ)
sel.register(p.stderr, selectors.EVENT_READ)

while True:
   for key, _ in sel.select():
       data = key.fileobj.read1(1024).decode()
       if not data:
           exit()
       data = data.strip()
       if data == "":
           continue
       if key.fileobj is p.stdout:
           logger.debug(data)
           print(data, end="")
       else:
           logger.error(data)
           print(data, end="", file=sys.stderr)

# URL:
# https://stackoverflow.com/questions/31833897/python-read-from-subprocess-stdout-and-stderr-separately-while-pr
# eserving-order
# Title: Python read from subprocess stdout and stderr separately while preserving order - Stack Overflow
"""
        )

        try:
            gdpass = subprocess.check_output(
                "echo NO_PASS", shell=True).decode("utf-8")
        except subprocess.CalledProcessError:
            gdpass = ""
        d = dict(
            gdrive_str=gdrive_str.replace(
                "CONTENT_CREDENTIAL", "" if os.getenv(
                    "CI") == "true" else gdpass
            ),
            rvs_pty_config_str=rvs_pty_config_str,
            rvs_str=rvs_str,
            runner_src=runner_src,
            AMQPURL=AMQPURL.string(),
            size=size,
            network=net,
            seed=seed,
            gdrive_enable=gdrive_enable,
            port=port,
            phase=phase,
            server=os.environ.get("server", "vtool.duckdns.org"),
        )
        ss = s.safe_substitute(d)

        with open(os.path.join(path, "main.py"), "w") as jf:
            jf.write(ss)

    def run_local(self, path):
        """run_local only for local test, real kernel runs in kaggle/colab server.

        :param path:
        """

        return subprocess.run("python " + os.path.join(path, "main.py"), shell=True)

    def create_runner(self, config, seed="2020", script=True, from_template=True):
        """create_runner.

        :param config: config will be size and model right now
        :param seed:
        :param script:
        :param from_template:
        """
        size = config["size"]
        net = config["network"]
        name = net.replace("_", "-") + "-" + str(size)
        AMQPURL = config["AMQPURL"]
        port = config["port"]
        phase = config.get("phase","dev")

        path = os.path.join(self.tmp_path, name)
        shutil.copytree(self.template_path, path)

        if from_template:
            self._change_kernel_meta_info(path, None, script)
            self._change_main_py(path, size, net, AMQPURL, seed, port, phase=phase)
        else:
            self._change_kernel_meta_info(
                path, self.title_prefix + " " + name, script)
            self._change_main_py(path, size, net, AMQPURL, seed, port, phase=phase)

        if not script:
            subprocess.run(
                ("jupytext --to notebook " + os.path.join(path, "main.py")).split()
            )

        self.runners.append(path)

        return path



import cv2
import numpy as np

from kaggle_runner.utils import kernel_utils


def post_process(probability, threshold, min_size):
    mask = cv2.threshold(probability, threshold, 1, cv2.THRESH_BINARY)[1]
    num_component, component = cv2.connectedComponents(mask.astype(np.uint8))
    predictions = np.zeros((1024, 1024), np.float32)
    num = 0
    for c in range(1, num_component):
        p = component == c
        if p.sum() > min_size:
            predictions[p] = 1
            num += 1
    return predictions, num


class PS_result_analyzer:  # todo maybe add other good analyze functions
    def dev_set_performance(self, y_true, y_pred):
        return kernel_utils.dice_coef(y_true, y_pred)


import os
import tensorflow as tf

from kaggle_runner import logger, may_debug
from kaggle_runner.defaults import DEBUG

AUTO = tf.data.experimental.AUTOTUNE

strategy = None
may_debug()
try:
    tpu_resolver = tf.distribute.cluster_resolver.TPUClusterResolver(os.getenv("TPU_NAME"))
    tf.config.experimental_connect_to_cluster(tpu_resolver)
    tf.tpu.experimental.initialize_tpu_system(tpu_resolver)
    strategy = tf.distribute.experimental.TPUStrategy(tpu_resolver)
    logger.debug('Running on TPU: %s', tpu_resolver.master())
except ValueError as e:
    logger.error("%s",e)
    tpu_resolver = None
    strategy = tf.distribute.get_strategy()
    logger.debug('NOT running on TPU.')


logger.info("REPLICAS: %d", strategy.num_replicas_in_sync)


EPOCHS = 2



import os
import shutil
import subprocess
from unittest import TestCase

import pytest
from kaggle_runner.runners import trainer
from kaggle_runner import utils


@pytest.fixture(scope="module")
def runner_configs():
    return [
        {"size": 384, "network": "intercept", "AMQPURL": utils.AMQPURL()},
        {"size": 384, "network": "intercept-resnet", "AMQPURL": utils.AMQPURL()},
    ]

class TestTrainer(TestCase):
    @classmethod
    def setup_class(cls):
        cls.tmp_path = ".runners"
        print("setup_class called once for the class")

    @classmethod
    def teardown_class(cls):
        print("teardown_class called once for the class")

    def setup_method(self, method):
        if os.path.exists(self.tmp_path):
            shutil.rmtree(self.tmp_path)
            os.mkdir(self.tmp_path)
        print("setup_method called for every method")

    def teardown_method(self, method):
        # shutil.rmtree(self.tmp_path)  # for debug
        print("teardown_method called for every method")
    def test_forward(self):
        self.fail()

    def test_iterate(self):
        self.fail()

    def test_start(self):
        self.fail()



import subprocess

from kaggle_runner import logger, may_debug

from kaggle_runner import defaults
defaults.LOAD_BERT_DATA = True
from kaggle_runner.datasets.bert import (BATCH_SIZE, TRAIN_LEN, train_dataset,
                                         val_data, valid_dataset, x_valid,
                                         y_valid)
from kaggle_runner.defaults import DEBUG
from kaggle_runner.utils.visualizer import visualize_model_preds

from kaggle_runner.kernels.bert import (bert_cbs,
                                        build_distilbert_model_singleton,
                                        get_test_result)


class Test_distilbert_model:
    @classmethod
    def setup_class(cls):
        # subprocess.run("make ripdbrv &", shell=True)

        if DEBUG:
            cls.model = build_distilbert_model_singleton(140)
        else:
            cls.model = build_distilbert_model_singleton(512)

    @classmethod
    def teardown_class(cls):
        #subprocess.run("pkill -f \"make ripdbrv\"", shell=True)
        try:
            del cls.model
        except Exception as e:
            print(e)
        logger.debug("tear down test %s", "Test_distilbert_model")

    def setup_method(self, method):
        logger.debug("setup for method %s", method)

    def teardown_method(self, method):
        logger.debug("teardown method %s", method)

    def test_data(self):
        assert val_data is not None
        assert train_dataset is not None

    def test_summary(self):
        self.setup_class()
        self.model.summary()
        assert True

    def test_fit_adv(self):
        # self.model_dev = build_distilbert_model_singleton(model_type="1st")

        if DEBUG:
            steps = 10
            epochs = 1
        else:
            steps = TRAIN_LEN//BATCH_SIZE
            logger.debug("Train len %s, batch size %s", TRAIN_LEN, BATCH_SIZE)
            epochs = 1
        logger.debug("Every epoch, steps is %s", steps)

        train_history = self.model.fit(
            train_dataset,
            steps_per_epoch=steps,
            validation_data=valid_dataset,
            callbacks=bert_cbs,
            epochs=epochs
        )
    def test_result(self):
        get_test_result(self)

    def test_visualize(self):
        # model_distilbert.summary()
        visualize_model_preds(self.model, val_data, x_valid, y_valid,
                              indices=[2,3, 5, 6, 7, 8, 1, 4])

if __name__ == "__main__":
    tt = Test_distilbert_model()
    tt.test_summary()
    tt.test_fit_adv()
    tt.test_visualize()
    tt.test_result()



import re
import random
import gc
import pandas as pd
import numpy as np
import gc
import cv2
import albumentations as A
from albumentations import (Blur, Compose, ElasticTransform, GaussNoise,
                            GridDistortion, HorizontalFlip, IAAEmboss,
                            Normalize, OneOf, #MultiplicativeNoise,
                            OpticalDistortion, RandomBrightnessContrast,
                            RandomGamma, RandomRotate90, Resize,
                            ShiftScaleRotate, Transpose, VerticalFlip)
from albumentations.pytorch import ToTensor
from albumentations.core.transforms_interface import DualTransform, BasicTransform
from kaggle_runner.utils.kernel_utils import get_obj_or_dump
from nltk import sent_tokenize

LANGS = {
    'en': 'english',
    'it': 'italian',
    'fr': 'french',
    'es': 'spanish',
    'tr': 'turkish',
    'ru': 'russian',
    'pt': 'portuguese'
}

class NLPTransform(BasicTransform):
    """ Transform for nlp task."""

    @property
    def targets(self):
        return {"data": self.apply}

    def update_params(self, params, **kwargs):
        if hasattr(self, "interpolation"):
            params["interpolation"] = self.interpolation

        if hasattr(self, "fill_value"):
            params["fill_value"] = self.fill_value

        return params

    def get_sentences(self, text, lang='en'):
        return sent_tokenize(text, LANGS.get(lang, 'english'))


class ShuffleSentencesTransform(NLPTransform):
    """ Do shuffle by sentence """

    def __init__(self, always_apply=False, p=0.5):
        super(ShuffleSentencesTransform, self).__init__(always_apply, p)

    def apply(self, data, **params):
        text, lang = data
        sentences = self.get_sentences(text, lang)
        random.shuffle(sentences)

        return ' '.join(sentences), lang


def get_open_subtitles():
    df_ot = get_pickled_data("ot.pkl")

    if df_ot is None:
        df_ot = pd.read_csv(f'{ROOT_PATH}/input/open-subtitles-toxic-pseudo-labeling/open-subtitles-synthesic.csv',
                            index_col='id')[['comment_text', 'toxic', 'lang']]
        df_ot = df_ot[~df_ot['comment_text'].isna()]
        df_ot['comment_text'] = df_ot.parallel_apply(
            lambda x: clean_text(x['comment_text'], x['lang']), axis=1)
        df_ot = df_ot.drop_duplicates(subset='comment_text')
        df_ot['toxic'] = df_ot['toxic'].round().astype(np.int)
        get_obj_or_dump("ot.pkl", default=df_ot)

    return df_ot

ROOT_PATH = '/kaggle'
def get_pickled_data(file_path):
    obj = get_obj_or_dump(file_path)

    if obj is None:
        #may_debug(True)

        return get_obj_or_dump(f"{ROOT_PATH}/input/clean-pickle-for-jigsaw-toxicity/{file_path}")

    return obj

class SynthesicOpenSubtitlesTransform(NLPTransform):
    def __init__(self, always_apply=False, supliment_toxic=None, p=0.5, mix=False):
        super(SynthesicOpenSubtitlesTransform, self).__init__(always_apply, p)

        df = get_open_subtitles()
        self.synthesic_toxic = df[df['toxic'] == 1].comment_text.values
        self.synthesic_non_toxic = df[df['toxic'] == 0].comment_text.values

        if supliment_toxic is not None:
            self.synthesic_toxic = np.concatenate(
                (self.synthesic_toxic, supliment_toxic))
        self.mix = mix

        del df
        gc.collect()

    def _mix_both(self, texts):
        for i in range(random.randint(0, 2)):
            texts.append(random.choice(self.synthesic_non_toxic))

        for i in range(random.randint(1, 3)):
            texts.append(random.choice(self.synthesic_toxic))

    def generate_synthesic_sample(self, text, toxic):
        texts = [text]

        if toxic == 0:
            if self.mix:
                self._mix_both(texts)
                toxic = 1
            else:
                for i in range(random.randint(1, 5)):
                    texts.append(random.choice(self.synthesic_non_toxic))
        else:
            self._mix_both(texts)
        random.shuffle(texts)

        return ' '.join(texts), toxic

    def apply(self, data, **params):
        text, toxic = data
        text, toxic = self.generate_synthesic_sample(text, toxic)

        return text, toxic


def get_transforms(phase, size, mean, std):
    list_transforms = []

    if phase == "train":
        list_transforms.extend(
            [
                HorizontalFlip(p=0.5),
                ShiftScaleRotate(
                    shift_limit=0,  # no resizing
                    scale_limit=0.1,
                    rotate_limit=10,  # rotate
                    p=0.5,
                    border_mode=cv2.BORDER_CONSTANT,
                ),
                GaussNoise(),
                #A.MultiplicativeNoise(multiplier=1.5, p=1),
            ]
        )
    list_transforms.extend(
        [Resize(size, size), Normalize(mean=mean, std=std, p=1), ToTensor(), ]
    )

    list_trfms = Compose(list_transforms)

    return list_trfms

# # + colab={} colab_type="code" id="8lcd0sZJ3bGj"


def get_sentences(text, lang='en'):
    return sent_tokenize(text, LANGS.get(lang, 'english'))

# # + colab={} colab_type="code" id="gMHg6Xvz3bGn"


def exclude_duplicate_sentences(text, lang='en'):
    sentences = []

    for sentence in get_sentences(text, lang):
        sentence = sentence.strip()

        if sentence not in sentences:
            sentences.append(sentence)

    return ' '.join(sentences)

# # + colab={} colab_type="code" id="EizVrxLZ3bGr"


def clean_text(text, lang='en'):
    text = str(text)
    text = re.sub(r'[0-9"]', '', text)
    text = re.sub(r'#[\S]+\b', '', text)
    text = re.sub(r'@[\S]+\b', '', text)
    text = re.sub(r'https?\S+', '', text)
    text = re.sub(r'\s+', ' ', text)
    text = exclude_duplicate_sentences(text, lang)

    return text.strip()


# # + colab={} colab_type="code" id="ffbjaiBH3bGu"
# # + colab={} colab_type="code" id="3b2yIDOv3bG1"


class ExcludeDuplicateSentencesTransform(NLPTransform):
    """ Exclude equal sentences """

    def __init__(self, always_apply=False, p=0.5):
        super(ExcludeDuplicateSentencesTransform,
              self).__init__(always_apply, p)

    def apply(self, data, **params):
        text, lang = data
        sentences = []

        for sentence in self.get_sentences(text, lang):
            sentence = sentence.strip()

            if sentence not in sentences:
                sentences.append(sentence)

        return ' '.join(sentences), lang

# # + colab={} colab_type="code" id="kyDSPoUF3bG4"


class ExcludeNumbersTransform(NLPTransform):
    """ exclude any numbers """

    def __init__(self, always_apply=False, p=0.5):
        super(ExcludeNumbersTransform, self).__init__(always_apply, p)

    def apply(self, data, **params):
        text, lang = data
        text = re.sub(r'[0-9]', '', text)
        text = re.sub(r'\s+', ' ', text)

        return text, lang

# # + colab={} colab_type="code" id="cAbXjyV63bG8"


class ExcludeHashtagsTransform(NLPTransform):
    """ Exclude any hashtags with # """

    def __init__(self, always_apply=False, p=0.5):
        super(ExcludeHashtagsTransform, self).__init__(always_apply, p)

    def apply(self, data, **params):
        text, lang = data
        text = re.sub(r'#[\S]+\b', '', text)
        text = re.sub(r'\s+', ' ', text)

        return text, lang

# # + colab={} colab_type="code" id="UQHHd_wo3bG_"


class ExcludeUsersMentionedTransform(NLPTransform):
    """ Exclude @users """

    def __init__(self, always_apply=False, p=0.5):
        super(ExcludeUsersMentionedTransform, self).__init__(always_apply, p)

    def apply(self, data, **params):
        text, lang = data
        text = re.sub(r'@[\S]+\b', '', text)
        text = re.sub(r'\s+', ' ', text)

        return text, lang

# # + colab={} colab_type="code" id="8sv-ecgw3bHC"


class ExcludeUrlsTransform(NLPTransform):
    """ Exclude urls """

    def __init__(self, always_apply=False, p=0.5):
        super(ExcludeUrlsTransform, self).__init__(always_apply, p)

    def apply(self, data, **params):
        text, lang = data
        text = re.sub(r'https?\S+', '', text)
        text = re.sub(r'\s+', ' ', text)

        return text, lang



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.4'
#       jupytext_version: 1.1.7
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# ## About this notebook
# We have seen much implementation of BERT and its varients (XLM RoBERTa , RoBERTa).So this leads me to try and experiment with other Language models .In this notebook i will try to provide a very concise use of GPT-2 model. You can change the model type from **Large , Xtralarge , Medium , DistilGPT-2**.
# So lets move.Also this note book is blatant copy of **xhlulu**'s kernel. so if you want to ***UPVOTE*** this kernel than be sure to ***UPVOTE*** his notebook too.
#
# ![](https://miro.medium.com/max/2880/0*BJTwVt0i59PaSB2Z.png)
#
# **THIS USE TRANSLATED DATA, AND IS TRAIN ON THE VALIDATION SET.**
#
#
# ### References
# * Original Author: [@xhlulu](https://www.kaggle.com/xhlulu/)
# * Original notebook: [Link](https://www.kaggle.com/xhlulu/jigsaw-tpu-distilbert-with-huggingface-and-keras)
#
# <font color="red" size=3>Please UPVOTE this kernel if you like it. It motivates me to produce more quality content :) <br><br>
#     Please do comment what your views are and what you understand from this.
#     <br><br> Dont't forget to UPVOTE xhlulu's notebook </font>

# + {"_cell_guid": "b1076dfc-b9ad-4769-8c92-a6c4dae69d19", "_uuid": "8f2839f25d086af736a60e9eeb907d3b93b6e0e5"}
from tensorflow.keras.callbacks import Callback
from sklearn.model_selection import StratifiedKFold
import warnings
import matplotlib.pyplot as plt
from tensorflow.keras import backend as K
from kaggle_runner.utils.kernel_utils import get_obj_or_dump
import re
from nltk import sent_tokenize
import os

import numpy as np
import pandas as pd
import tensorflow as tf
from tensorflow.keras.layers import Dense, Input
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.models import Model
from tensorflow.keras.callbacks import ModelCheckpoint
#from kaggle_datasets import KaggleDatasets
import transformers
from transformers import TFAutoModel, AutoTokenizer
from tqdm.notebook import tqdm
from keras.preprocessing import sequence
from tokenizers import Tokenizer, models, pre_tokenizers, decoders, processors


# -

# ## Helper Functions

# + {"_cell_guid": "79c7e3d0-c299-4dcb-8224-4455121ee9b0", "_uuid": "d629ff2d2480ee46fbb7e2d37f6b5fab8052498a"}
def fast_encode(texts, tokenizer, chunk_size=256, maxlen=192):
    """
    https://www.kaggle.com/xhlulu/jigsaw-tpu-distilbert-with-huggingface-and-keras
    """
    tokenizer.enable_truncation(max_length=maxlen)
    tokenizer.enable_padding(max_length=maxlen)
    all_ids = []

    for i in tqdm(range(0, len(texts), chunk_size)):
        text_chunk = texts[i:i+chunk_size].tolist()
        encs = tokenizer.encode_batch(text_chunk)
        all_ids.extend([enc.ids for enc in encs])

    return np.array(all_ids)


# -

def gpt2_encode(texts, tokenizer, maxlen=192):
    enc_di = tokenizer.batch_encode_plus(
        texts,
        return_attention_masks=False,
        return_token_type_ids=False,
        max_length=maxlen,
    )
    pad_idx = tokenizer.convert_tokens_to_ids(['<pad>'])[0]
    enc_di['input_ids'] = sequence.pad_sequences(
        enc_di['input_ids'], maxlen=MAX_LEN, padding='post', value=pad_idx)

    return np.array(enc_di['input_ids'])


# ## TPU Configs

# +
# Detect hardware, return appropriate distribution strategy
try:
    # TPU detection. No parameters necessary if TPU_NAME environment variable is
    # set: this is always the case on Kaggle.
    tpu = tf.distribute.cluster_resolver.TPUClusterResolver()
    print('Running on TPU ', tpu.master())
except ValueError:
    tpu = None

if tpu:
    tf.config.experimental_connect_to_cluster(tpu)
    tf.tpu.experimental.initialize_tpu_system(tpu)
    strategy = tf.distribute.experimental.TPUStrategy(tpu)
else:
    # Default distribution strategy in Tensorflow. Works on CPU and single GPU.
    strategy = tf.distribute.get_strategy()

print("REPLICAS: ", strategy.num_replicas_in_sync)
# -

# ## Create fast tokenizer

# ## Load text data into memory

os.listdir('/kaggle/input')

# !cp /kaggle/input/bert-for-toxic-classfication-trained/* .

AUTO = tf.data.experimental.AUTOTUNE

# DGCS_DS_PATHata access
#GCS_DS_PATH = KaggleDatasets().get_gcs_path(
#    'jigsaw-multilingual-toxic-comment-classification')
#MY_GCS_DS_PATH = KaggleDatasets().get_gcs_path('jigsaw-new-balanced-dataset')
#NEW_GCS_DS_PATH = KaggleDatasets().get_gcs_path(
#    'jigsaw-multilingual-toxic-test-translated')

# Configuration
EPOCHS = 5
BATCH_SIZE = 16 * strategy.num_replicas_in_sync
MAX_LEN = 192

# +
LANGS = {
    'en': 'english',
    'it': 'italian',
    'fr': 'french',
    'es': 'spanish',
    'tr': 'turkish',
    'ru': 'russian',
    'pt': 'portuguese'
}


def get_sentences(text, lang='en'):
    return sent_tokenize(text, LANGS.get(lang, 'english'))


def exclude_duplicate_sentences(text, lang='en'):
    sentences = []

    for sentence in get_sentences(text, lang):
        sentence = sentence.strip()

        if sentence not in sentences:
            sentences.append(sentence)

    return ' '.join(sentences)


def clean_text(text, lang='en'):
    text = str(text)
    text = re.sub(r'[0-9"]', '', text)
    text = re.sub(r'#[\S]+\b', '', text)
    text = re.sub(r'@[\S]+\b', '', text)
    text = re.sub(r'https?\S+', '', text)
    text = re.sub(r'\s+', ' ', text)

    return text.strip()


# -

MODEL = 'gpt2-medium'
tokenizer = AutoTokenizer.from_pretrained(MODEL)

# !git clone https://github.com/pennz/kaggle_runner k

# !rsync -r  k/* .
# !python3 -m pip install -e .

# +
X, y, X_valid, Y_valid, X_test = None, None, None, None, None
y = get_obj_or_dump("y.pkl", default=y)
X = get_obj_or_dump("x.pkl", default=X)


X_valid = get_obj_or_dump("xval.pkl", default=X_valid)
Y_valid = get_obj_or_dump("yval.pkl", default=Y_valid)
# -

X_test = get_obj_or_dump("xtest.pkl", default=X_test)

len(X), len(y)

# +


def focal_loss(gamma=2., alpha=.2):
    def focal_loss_fixed(y_true, y_pred):
        pt_1 = tf.where(tf.equal(y_true, 1), y_pred, tf.ones_like(y_pred))
        pt_0 = tf.where(tf.equal(y_true, 0), y_pred, tf.zeros_like(y_pred))

        return -K.mean(alpha * K.pow(1. - pt_1, gamma) * K.log(pt_1)) - K.mean((1 - alpha) * K.pow(pt_0, gamma) * K.log(1. - pt_0))

    return focal_loss_fixed


# +
def build_lrfn(lr_start=0.000001, lr_max=0.000004,
               lr_min=0.0000001, lr_rampup_epochs=7,
               lr_sustain_epochs=0, lr_exp_decay=.87):
    lr_max = lr_max * strategy.num_replicas_in_sync

    def lrfn(epoch):
        if epoch < lr_rampup_epochs:
            lr = (lr_max - lr_start) / lr_rampup_epochs * epoch + lr_start
        elif epoch < lr_rampup_epochs + lr_sustain_epochs:
            lr = lr_max
        else:
            lr = (lr_max - lr_min) * lr_exp_decay**(epoch -
                                                    lr_rampup_epochs - lr_sustain_epochs) + lr_min

        return lr

    return lrfn


# +
plt.figure(figsize=(10, 7))

_lrfn = build_lrfn()
plt.plot([i for i in range(35)], [_lrfn(i) for i in range(35)])


# -

def build_model(transformer, loss, max_len=192):

    input_word_ids = Input(
        shape=(max_len,), dtype=tf.int32, name="input_word_ids")
    sequence_output = transformer(input_word_ids)[0]
    cls_token = sequence_output[:, 0, :]
    out = Dense(1, activation='sigmoid')(cls_token)

    model = Model(inputs=input_word_ids, outputs=out)
    model.compile(Adam(lr=1e-6), loss=loss, metrics=[tf.keras.metrics.AUC()])

    return model


with strategy.scope():
    transformer_layer = TFAutoModel.from_pretrained(MODEL)
    model = build_model(transformer_layer,
                        loss='binary_crossentropy', max_len=MAX_LEN)

# %%time
warnings.filterwarnings('ignore')
pred = []
skf = StratifiedKFold(n_splits=3, shuffle=True, random_state=42)
fold = 1

for train_index, test_index in skf.split(X, y):
    X_train, a = X[train_index], X[test_index]
    Y_train, v = y[train_index], y[test_index]

    train_dataset = (
        tf.data.Dataset
        .from_tensor_slices((X_train, Y_train))
        .repeat()
        .shuffle(2048)
        .batch(BATCH_SIZE)
        .prefetch(AUTO)
    )

    valid_dataset = (
        tf.data.Dataset
        .from_tensor_slices((X_valid, Y_valid))
        .batch(BATCH_SIZE)
        .prefetch(AUTO)
    )

    test_dataset = (
        tf.data.Dataset
        .from_tensor_slices(X_test)
        .batch(BATCH_SIZE)
        .prefetch(AUTO)
    )

    lrfn = build_lrfn()
    lr_schedule = tf.keras.callbacks.LearningRateScheduler(lrfn, verbose=1)
    n_steps = X_train.shape[0] // BATCH_SIZE

    train_history = model.fit(
        train_dataset,
        steps_per_epoch=n_steps,
        validation_data=valid_dataset,
        callbacks=[lr_schedule],
        epochs=EPOCHS
    )
    sub['toxic'] = model.predict(test_dataset, verbose=1)
    sub.to_csv(str(fold) + 'gpt2m_submission.csv', index=False)
    fold = fold + 1
    print(fold)

# # FINE

# ```python
# train1 = pd.read_csv("/kaggle/input/jigsaw-multilingual-toxic-comment-classification/jigsaw-toxic-comment-train.csv")
# train2 = pd.read_csv("/kaggle/input/jigsaw-multilingual-toxic-comment-classification/jigsaw-unintended-bias-train.csv")
# train2.toxic = train2.toxic.round().astype(int)
# train3 = pd.read_csv("/kaggle/input/jigsaw-new-balanced-dataset/new_toxic_comments.csv")
# train3['toxic'] = 1
# a=set(train1.columns)
#
# b = a.intersection(train2.columns)
#
# # Combine train1 with a subset of train2
# train = pd.concat([
#     train1[b],
#     train2[b].query('toxic==1'),
#     train2[b].query('toxic==0').sample(n=300000, random_state=42)
# ])
# del train1
# del train2
#
# train = train.drop(columns=['id'])
# train = train[['comment_text','toxic']]
# train3.columns = train.columns
# train = pd.concat([train,train3],axis=0)
# train['comment_text']  = train.comment_text.apply(lambda x : clean_text(x))
# from sklearn.utils import shuffle
# train = shuffle(train).reset_index(drop=True)
#
#
# X_valid =  pd.read_csv("/kaggle/input/jigsaw-multilingual-toxic-comment-classification/validation.csv")
# Y_valid = X_valid.toxic.values
# X_valid = gpt2_encode(X_valid.comment_text.values, tokenizer, maxlen=MAX_LEN)
#
# X_test = pd.read_csv("/kaggle/input/jigsaw-multilingual-toxic-comment-classification/test.csv")
# sub = pd.read_csv("/kaggle/input/jigsaw-multilingual-toxic-comment-classification/test.csv")
# X_test = gpt2_encode(X_test.content.values, tokenizer, maxlen=MAX_LEN)
#
#
# X = gpt2_encode(train.comment_text.values, tokenizer, maxlen=MAX_LEN)
# y = train.toxic.values
# del train
# import gc
# gc.collect()
# ```



# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: id,colab_type,colab,language,-all
#     formats: ipynb,py
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.4'
#       jupytext_version: 1.1.7
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# + {"id": "view-in-github", "colab_type": "text", "cell_type": "markdown"}
# <a href="https://colab.research.google.com/github/pennz/kaggle_runner/blob/master/hub/shonenkov_training_pipeline.ipynb" target="_parent"><img src="https://colab.research.google.com/assets/colab-badge.svg" alt="Open In Colab"/></a>

# + {"colab_type": "code", "id": "-5g67DNAUQyh", "colab": {}}
# %load_ext autoreload
# %autoreload 2

# + {"colab_type": "code", "id": "n-Gf0MyQL8Ys", "colab": {}}
#@title Kaggle thing

text = "{   \"username\": \"k1gaggle\",   \"key\": \"721ad312727847f609212568cf015532\",   \"competition\": \"siim-acr-pneumothorax-segmentation\" }" #@param {type:"string"}
#dropdown = '1st option' #@param ["1st option", "2nd option", "3rd option"]
#text_and_dropdown = 'value' #@param ["1st option", "2nd option", "3rd option"] {allow-input: true}
# !mkdir -p ~/.kaggle/
with open('/root/.kaggle/kaggle.json', 'w') as f:
    f.write(text)
#print(dropdown)
#print(text_and_dropdown)

# + {"colab_type": "code", "id": "A6Yz0kroUQy0", "colab": {"base_uri": "https://localhost:8080/", "height": 34}, "magic_args": "--bg", "language": "bash"}
# #python3 -m pip show kaggle_runner || ( git clone https://github.com/pennz/kaggle_runner; \
# #mv kaggle_runner k && \
# #mv k/* . && mv k/.* .; \
# #python3 -m pip install -e .; \
# #git submodule update --init; \
# #export PATH=$PWD/bin:$PATH; \
# #entry.sh; echo You can wait to setup for remote access)


# + {"colab_type": "code", "id": "udc5hHgUUQzD", "colab": {"base_uri": "https://localhost:8080/", "height": 34}}
import subprocess
subprocess.run("""python3 -m pip show kaggle_runner || ( git clone https://github.com/pennz/kaggle_runner;
mv kaggle_runner k && mv k/* . && mv k/.* .;
python3 -m pip install -e .;
git submodule update --init;
export PATH=$PWD/bin:$PATH; entry.sh &
echo You can wait to setup for remote access)
""", shell=True)

import subprocess
subprocess.run("make install_dep; mkdir -p /root/.ssh ; make kr; wait; make gitlab; make xla &", shell=True)


# + {"colab_type": "code", "id": "9aQY9GJrVmGl", "colab": {"base_uri": "https://localhost:8080/", "height": 121}}
# #!yes | make vim &

# + {"colab_type": "code", "id": "6lwyNn2ONmLR", "colab": {"base_uri": "https://localhost:8080/", "height": 1000}}
# !make dd

# + {"colab_type": "code", "id": "3ccg5MJlyvLy", "colab": {"base_uri": "https://localhost:8080/", "height": 1000}}
# !make install_dep

# + {"colab_type": "code", "id": "MYBRID_uUQzb", "colab": {"base_uri": "https://localhost:8080/", "height": 159}}
from importlib import reload
import kaggle_runner
reload(kaggle_runner)
from kaggle_runner import may_debug, logger
from kaggle_runner.modules.ToxicSimpleNNModel import ToxicSimpleNNModel
from kaggle_runner.kernels.Shonenkov import Shonenkov, ShonenkovChangeInner
from kaggle_runner.callbacks import CheckGrad,_check_grad
from kaggle_runner.metrics.meters import AverageMeter, RocAucMeter
from kaggle_runner.losses import LabelSmoothing
from kaggle_runner.datasets.transfomers import *
from kaggle_runner import defaults


# + {"colab_type": "code", "id": "bQfQoj4PUQzl", "colab": {}}
import numpy as np
import pandas as pd
import os
os.environ['XLA_USE_BF16'] = "1"


# + {"colab_type": "code", "id": "jryGiXHWUQzp", "colab": {}}
from glob import glob

# + {"colab_type": "code", "id": "0qyEzTQDUQzu", "colab": {}}



# + {"colab_type": "code", "id": "njd0Gd2TUQz0", "colab": {}}
import torch
import torch.nn as nn
from torch.utils.data import Dataset,DataLoader
from torch.autograd import Variable
from torch.utils.data.sampler import SequentialSampler, RandomSampler
import sklearn


# + {"colab_type": "code", "id": "0k4-gbQhUQz5", "colab": {}}
import time
import random
from datetime import datetime
from tqdm import tqdm
tqdm.pandas()


# + {"colab_type": "code", "id": "63pHGleDUQ0D", "colab": {}}
import fastai
from fastai import *
from fastai.core import *
from fastai.torch_core import *
from fastai.vision import *
from fastai.callbacks.misc import StopAfterNBatches
from fastai.callbacks import *
from transformers import AdamW, get_linear_schedule_with_warmup, get_constant_schedule
from fastai.text.transform import Vocab


# + {"colab_type": "code", "id": "MMbcpmJHUQ0H", "colab": {}}
import gc
import re


# + {"colab_type": "code", "id": "-NTVjOreUQ0L", "colab": {"base_uri": "https://localhost:8080/", "height": 69}}
# # !python3 -m pip install nltk > /dev/null
import nltk
nltk.download('punkt')

# + {"colab_type": "code", "id": "EfgDxZSiUQ0O", "colab": {}}
from nltk import sent_tokenize

# + {"colab_type": "code", "id": "JpTc5nfQUQ0T", "colab": {"base_uri": "https://localhost:8080/", "height": 52}}
from pandarallel import pandarallel
pandarallel.initialize(nb_workers=4, progress_bar=False)

# + {"colab_type": "code", "id": "itAbsNWXUQ0W", "colab": {}}
import warnings
warnings.filterwarnings("ignore")


# + {"colab_type": "code", "id": "JHro74jvUQ0d", "colab": {}}
ROOT_PATH = f'/kaggle' # for colab

# + {"colab_type": "code", "id": "IJD6tJMaUQ0i", "colab": {}}
def get_toxic_comments(df):
    df = df[~df['comment_text'].isna()]
    df = df.drop_duplicates(subset='comment_text')
    df['toxic'] = df['toxic'].round().astype(np.int)

    return df[df['toxic'] == 1].comment_text.values


# + {"colab_type": "code", "id": "MpvJJZsbUQ0n", "colab": {"base_uri": "https://localhost:8080/", "height": 34}}
# #![ -f train.pkl ] || cp /kaggle/input/clean-pickle-for-jigsaw-toxicity/*pkl .
subprocess.run('[ -f train.pkl ] || cp /kaggle/input/clean-pickle-for-jigsaw-toxicity/*pkl .', shell=True)


# + {"colab_type": "code", "id": "cmkH1Q0qUQ0t", "colab": {}}
class TrainGlobalConfig:
    """ Global Config for this notebook """
    num_workers = 0  # количество воркеров для loaders
    batch_size = 16  # bs , 8 for GPU, 16 for TPU
    n_epochs = 2  # количество эпох для обучения
    lr = 0.5 * 1e-5 # стартовый learning rate (внутри логика работы с мульти TPU домножает на кол-во процессов)
    fold_number = 0  # номер фолда для обучения

    # -------------------
    verbose = True  # выводить принты
    verbose_step = 25  # количество шагов для вывода принта
    # -------------------

    # --------------------
    step_scheduler = False  # выполнять scheduler.step после вызова optimizer.step
    validation_scheduler = True  # выполнять scheduler.step после валидации loss (например для плато)
    SchedulerClass = torch.optim.lr_scheduler.ReduceLROnPlateau
    scheduler_params = dict(
        mode='max',
        factor=0.7,
        patience=0,
        verbose=False,
        threshold=0.0001,
        threshold_mode='abs',
        cooldown=0,
        min_lr=1e-8,
        eps=1e-08
    )
    # --------------------

    # -------------------
    criterion = LabelSmoothing()
    # -------------------

# + {"colab_type": "code", "id": "pB36YxahXT6Q", "colab": {}}
from kaggle_runner.kernels.Shonenkov import Shonenkov

# + {"colab_type": "code", "id": "furzEDtzUQ0x", "colab": {"base_uri": "https://localhost:8080/", "height": 321, "referenced_widgets": ["ee035e7ededd4627bd0c945937717ade", "e327a32093be48ab9fabd6d2f93d4c0a", "8533de2bc7134b40abacf7a9b32f0012", "9c06456568914165afb73ac30228ba06", "5ba48862ccdf438c856405814423029e", "aa4533e1247846fcbf408a8be3a1359d", "00bdb7e7c8a0450da3d432a057de1ba1", "e09d32adcc504e85a050a2fd8a8ab55c", "ff36006e3f57454fa52fb7bf1647c21a", "4218600849cd4b2c8573917cf4e7b0cc", "9e062caa49984f7e9773e6d74a7cfadf", "552da08823e04c6693f3547e0e63f5cc", "5d534c0d831b4c3e8b25229ff5949366", "94309ba4763e4a74b447c70f15fd8e07", "88309c98031a4fbdbc0e247de4e70e4a", "e7bc1adc51564d86ae8be192335b184c", "72a4d2af4a164217843d68c79e473b6c", "e5783b109ca6436b8bb62f1b8a49706f", "68eb7fcbb8c544fcad5c271a10b2a303", "1098a4a2ecc84ac39c5ce37904f2a75d", "c6655ec11ee44c35be4ed66fec0bd0df", "ed53c651dc9d4ec598feff58b65fde84", "93ea82bc7a7545e29ad646a5f83b1358", "627f81a4d35742bba995678bb36436ad"]}}
k = Shonenkov(torch.device("cpu"), TrainGlobalConfig, metrics=None, loss_func=LabelSmoothing(), opt_func=None)
k.run(dump_flag=False)

# + {"colab_type": "code", "id": "um8I7KBrY0Ur", "colab": {}}
subprocess.run('mkdir ./models_xlmrobert/', shell=True)

# + {"colab_type": "code", "id": "gDt1SW3AUQ02", "colab": {}}
from transformers import WEIGHTS_NAME, CONFIG_NAME
def save_model(self, output_dir="./models_xlmrobert/"):
    subprocess.run(f'mkdir {output_dir}', shell=True)
    model = self.model

# Step 1: Save a model, configuration and vocabulary that you have fine-tuned

# If we have a distributed model, save only the encapsulated model
# (it was wrapped in PyTorch DistributedDataParallel or DataParallel)
    model_to_save = model.module if hasattr(model, 'module') else model

# If we save using the predefined names, we can load using `from_pretrained`
    output_model_file = os.path.join(output_dir, WEIGHTS_NAME)
    output_config_file = os.path.join(output_dir, CONFIG_NAME)

    torch.save(model_to_save.state_dict(), output_model_file)
    model_to_save.backbone.config.to_json_file(output_config_file)
    #tokenizer.save_pretrained(output_dir)

# + {"colab_type": "code", "id": "PN07kHKOUQ06", "colab": {}}
from catalyst.data.sampler import DistributedSamplerWrapper, BalanceClassSampler
def test_model_fn(device=torch.device("cpu")):
    #device = xm.xla_device(devkind='TPU')
    #device=torch.device("xla")
    logger.debug("Device used: %s", device)

    #k.run(dump_flag=True) # it seems it cannot save right
    #k.run(dump_flag=False)
    #k.peek_data()

    self = k
    assert self.validation_dataset is not None
    #assert self.learner is not None

    net = k.model
    assert net is not None
    net.to(device)

    param_optimizer = list(self.model.named_parameters())
    no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
    optimizer_grouped_parameters = [
        {'params': [p for n, p in param_optimizer if not any(nd in n for nd in no_decay)], 'weight_decay': 0.001},
        {'params': [p for n, p in param_optimizer if any(nd in n for nd in no_decay)], 'weight_decay': 0.0}
    ]
    #optimizer = AdamW(optimizer_grouped_parameters, lr=TrainGlobalConfig.lr*xm.xrt_world_size())
    optimizer = AdamW(optimizer_grouped_parameters, lr=TrainGlobalConfig.lr*8)

    train_loader = torch.utils.data.DataLoader(
        self.train_dataset,
        batch_size=TrainGlobalConfig.batch_size,
        shuffle=False, # sampler is set, so shuffle here should be False
        sampler=BalanceClassSampler(labels=k.train_dataset.get_labels(), mode="downsampling"),
        pin_memory=False,
        drop_last=True,
        num_workers=TrainGlobalConfig.num_workers,
    )
    validation_loader = torch.utils.data.DataLoader(
        self.validation_dataset,
        batch_size=TrainGlobalConfig.batch_size,
    #    sampler=validation_sampler,
        pin_memory=False,
        drop_last=False,
        num_workers=TrainGlobalConfig.num_workers
    )
    test_loader = torch.utils.data.DataLoader(
        self.test_dataset,
        batch_size=TrainGlobalConfig.batch_size,
    #    sampler=test_sampler,
        pin_memory=False,
        drop_last=False,
        num_workers=TrainGlobalConfig.num_workers
    )
    validation_tune_loader = torch.utils.data.DataLoader(
        self.validation_tune_dataset,
        batch_size=TrainGlobalConfig.batch_size,
        #sampler=validation_tune_sampler,
        pin_memory=False,
        drop_last=False,
        num_workers=TrainGlobalConfig.num_workers
    )

    def validation(model, device, config, val_loader, criterion):
        model.eval()
        losses = AverageMeter()
        final_scores = RocAucMeter()

        t = time.time()

        for step, (inputs_masks, targets) in enumerate(val_loader):
            inputs=inputs_masks[0]
            attention_masks=inputs_masks[1]

            if config.verbose:
                if step % config.verbose_step == 0:
                    logger.info(
                        f'Valid Step {step}, loss: ' + \
                        f'{losses.avg:.5f}, final_score: {final_scores.avg:.5f}, mc_score: {final_scores.mc_avg:.5f}, ' + \
                        f'time: {(time.time() - t):.5f}'
                    )
            with torch.no_grad():
                inputs = inputs.to(device, dtype=torch.long)
                attention_masks = attention_masks.to(device, dtype=torch.long)
                targets = targets.to(device, dtype=torch.float)

                outputs = model(inputs, attention_masks)
                loss = criterion(outputs, targets)

                batch_size = inputs.size(0)

                final_scores.update(targets, outputs)
                losses.update(loss.detach().item(), batch_size)

    def run_inference(model, device, config, test_loader):
        model.eval()
        result = {'id': [], 'toxic': []}
        t = time.time()

        for step, (inputs_masks, ids) in enumerate(test_loader):
            inputs=inputs_masks[0]
            attention_masks=inputs_masks[1]

            if config.verbose:
                if step % config.verbose_step == 0:
                    logger.info(f'Prediction Step {step}, time: {(time.time() - t):.5f}')

            with torch.no_grad():
                inputs = inputs.to(device, dtype=torch.long)
                attention_masks = attention_masks.to(device, dtype=torch.long)
                outputs = model(inputs, attention_masks)
                toxics = nn.functional.softmax(outputs, dim=1).data.cpu().numpy()[:,1]

            result['id'].extend(ids.cpu().numpy())
            result['toxic'].extend(toxics)

            break # just test one batch

        return result

    def train_one_epoch(model, device, config, train_loader, criterion, optimizer):
        model.train()

        losses = AverageMeter()
        final_scores = RocAucMeter()
        t = time.time()

        for step, (inputs_masks, targets) in enumerate(train_loader):
            inputs=inputs_masks[0]
            attention_masks=inputs_masks[1]

            batch_size = inputs.size(0)

            if config.verbose:
                if step % config.verbose_step == 0:
                    logger.debug(
                        f'Train Step {step}, bs: {batch_size}, loss: ' + \
                        f"{losses.avg:.5f}, lr: {optimizer.param_groups[0]['lr']} final_score: {final_scores.avg:.5f}, mc_score: {final_scores.mc_avg:.5f}, " + \
                        f'time: {(time.time() - t):.5f}'
                    )

            inputs = inputs.to(device, dtype=torch.long)
            attention_masks = attention_masks.to(device, dtype=torch.long)
            targets = targets.to(device, dtype=torch.float)

            optimizer.zero_grad()

            outputs = model(inputs, attention_masks)
            loss = criterion(outputs, targets)


            final_scores.update(targets, outputs)

            losses.update(loss.detach().item(), batch_size)

            loss.backward()
            _check_grad(optimizer)
            #optimizer.step()
            xm.optimizer_step(optimizer, barrier=True)

        model.eval()
        #self.save('last-checkpoint.bin')

        return losses, final_scores

    def run_tuning_and_inference(net, device, TrainGlobalConfig, validation_loader, train_loader):
        for e in range(1):
            self.optimizer.param_groups[0]['lr'] = self.config.lr*8

            losses, final_scores = train_one_epoch(net, device, TrainGlobalConfig, train_loader, TrainGlobalConfig.criterion, )
            self.log(f'[RESULT]: Tune_Train. Epoch: {self.epoch}, loss: {losses.avg:.5f}, final_score: {final_scores.avg:.5f}, mc_score: {final_scores.mc_avg:.5f}, time: {(time.time() - t):.5f}')

            t = time.time()
            para_loader = pl.ParallelLoader(validation_loader, [self.device])
            losses, final_scores = self.validation(para_loader.per_device_loader(self.device))
            self.log(f'[RESULT]: Tune_Validation. Epoch: {self.epoch}, loss: {losses.avg:.5f}, final_score: {final_scores.avg:.5f}, mc_score: {final_scores.mc_avg:.5f}, time: {(time.time() - t):.5f}')

            run_inference(net, device, TrainGlobalConfig, validation_loader)

    #train_one_epoch(net, device, TrainGlobalConfig, train_loader, TrainGlobalConfig.criterion, optimizer)
    #losses, final_scores = validation(net, device, TrainGlobalConfig, validation_loader, TrainGlobalConfig.criterion)
    #logger.info(f"Val results: losses={losses}, final_scores={final_scores}")

    results = run_inference(net, device, TrainGlobalConfig, validation_loader)
    logger.info(f"Test done, result len %d", len(results))

# + {"colab_type": "code", "id": "0qUwvlFnUQ0-", "colab": {}}
def load_model(self, output_dir="./models_xlmrobert/"):
    output_model_file = os.path.join(output_dir, WEIGHTS_NAME)
    state_dict = torch.load(output_model_file)
    self.model.load_state_dict(state_dict)

    print(self.model)

def test_save_model():
    save_model(k)

def test_load():
    load_model(k)

# + {"colab_type": "code", "id": "9dqdFgwAZQgw", "colab": {"base_uri": "https://localhost:8080/", "height": 1000}}
test_save_model()
test_load()
test_model_fn()

import pytest
import sys

def test_exit():
    with pytest.raises(SystemExit) as exc:
        # test goes here

        if os.getenv("CI") == "true":
            sys.exit(1)
        else:
            sys.exit(0)
    assert exc.value.code == 0

test_exit()

# + {"colab_type": "code", "id": "NSC6BrbwUQ2G", "colab": {}}

if os.getenv("CI") != "true":

# + {"colab_type": "code", "id": "VXNH-qcCUQ1B", "colab": {}}
    from kaggle_runner.kernels.fastai_kernel import FastAIKernel
    from kaggle_runner.runners.trainer import GPUTrainer
    def _to_gpu(learn:Learner, k: FastAIKernel) -> Learner:
        learn.callback_fns.append(partial(GPUTrainer, k=k))

        return learn

    Learner.to_gpu = _to_gpu


# + {"colab_type": "code", "id": "2Z1O-ZLfUQ1D", "colab": {}}
    import pysnooper
    from functools import partial

    from hub.custom_fastai_callbacks.callbacks import GradientAccumulator
    def debug_train(use_dist_cb=True):
        logger.debug(f'debug train with{" " if use_dist_cb else "OUT"} to_tpu_distributed')
        from kaggle_runner import defaults
        _DEBUG = defaults.DEBUG
        #defaults.DEBUG = True

        param_optimizer = list(k.model.named_parameters())
        no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
        optimizer_grouped_parameters = [
            {'params': [p for n, p in param_optimizer if not any(nd in n for nd in no_decay)], 'lr': 0., 'weight_decay': 0.01},
            {'params': [p for n, p in param_optimizer if any(nd in n for nd in no_decay)], 'lr': 0., 'weight_decay': 0.0}
        ]

        def AdamW_with_given_p(p_to_ignore, *args, **kargs):
            kargs['lr']=TrainGlobalConfig.lr*8 #xm.xrt_world_size()

            return AdamW(optimizer_grouped_parameters, *args, **kargs)

        learn = k.create_learner(k, opt_func=AdamW_with_given_p,
                                 loss_func=LabelSmoothing(),
                                 wd=0.01,
                                 callback_fns=[partial(GradientClipping, clip=0.5),
                                               partial(CSVLogger, append=True),
                                               partial(GradientAccumulator, num_iterations=4),
                                               partial(CheckGrad, skip_loss_step=False, batch_size=k.config.batch_size)]
                                 )
        k.learner = learn

        if use_dist_cb:
            learn = learn.to_tpu_distributed()
        else:
            learn = learn.to_gpu(k)

        #learn.callback_fns.append(CheckGrad)
        #print('hello')
        #learn.lr_find(start_lr=1e-7, end_lr=1e-2, num_it=200)
        #learn.recorder.plot()
        learn.fit_one_cycle(2, max_lr=3e-5)
        #learn.fit(1, lr=4e-5) # original 0.5*e-5*8=4*e-5
        defaults.DEBUG = _DEBUG


# + {"colab_type": "code", "id": "uUzCsw_wUQ1F", "colab": {}}
# %%time
#debug_train(use_dist_cb=False)


# + {"colab_type": "text", "id": "rO2ay06LUQ1I", "cell_type": "markdown"}
# # XLA


# + {"colab_type": "code", "id": "eeL1PlTnUQ1J", "colab": {}}
    import subprocess
    subprocess.run('make xla', shell=True) # TODO add system call, we can get log?

# + {"colab_type": "code", "id": "5fUVZUIYUQ1M", "colab": {}}
    import torch_xla
    import torch_xla.distributed.data_parallel as dp
    import torch_xla.utils.utils as xu
    import torch_xla.core.xla_model as xm
    import torch_xla.distributed.parallel_loader as pl
    import torch_xla.distributed.xla_multiprocessing as xmp
    import torch

# + {"colab_type": "code", "id": "qWaNMw-rUQ1O", "colab": {}}
    import fastai
    from fastai import *
    from fastai.core import *
    from fastai.torch_core import *
    from fastai.vision import *
    from fastai.basic_train import *
    from kaggle_runner.runners.tpu_trainer import TPUDistributed, TPUFitter

# + {"colab_type": "code", "id": "VomH5RztUQ1T", "colab": {}}
    from catalyst.data.sampler import DistributedSamplerWrapper, BalanceClassSampler
    def len_parallelloader(self):
        return len(self._loader._loader)
    pl.PerDeviceLoader.__len__ = len_parallelloader


# + {"colab_type": "code", "id": "0d8Bzl8DUQ1X", "colab": {}}
    def _to_tpu_distributed(learn:Learner) -> Learner:
        learn.callback_fns.append(TPUDistributed)

        return learn

# + {"colab_type": "code", "id": "a6_DgUIyUQ1Z", "colab": {}}
    Learner.to_tpu_distributed = _to_tpu_distributed


# + {"colab_type": "code", "id": "M_tjbI16UQ1g", "colab": {}}
    from functools import partial
    import pysnooper

# + {"colab_type": "code", "id": "0nejQ0FwUQ1j", "colab": {}}
    @pysnooper.snoop()
    def train_loop(index, *args):
        logger.debug("rank: %d entered train_loop", index)

        param_optimizer = list(k.model.named_parameters())
        no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
        optimizer_grouped_parameters = [
            {'params': [p for n, p in param_optimizer if not any(nd in n for nd in no_decay)], 'lr': 4e-5, 'weight_decay': 0.01},
            {'params': [p for n, p in param_optimizer if any(nd in n for nd in no_decay)], 'lr': 4e-5, 'weight_decay': 0.0}
        ]

        def AdamW_with_given_p(p_to_ignore, *args, **kargs):
            kargs['lr']=TrainGlobalConfig.lr*xm.xrt_world_size()

            return AdamW(optimizer_grouped_parameters, *args, **kargs)

        if index == 0:
            time.sleep(1)
        learn = k.create_learner(k, opt_func=AdamW_with_given_p,
                                 loss_func=LabelSmoothing(),
                                 wd=0.01,
                                 callback_fns=[partial(GradientClipping, clip=0.5),
                                               ShowGraph,
                                               partial(CSVLogger, append=True),
                                               partial(CheckGrad, skip_loss_step=False)]
                                 ).to_tpu_distributed()
        learn.lr_find(start_lr=1e-7, end_lr=1e-5, num_it=200)
        learn.recorder.plot()
        #learn.fit_one_cycle(3, max_lr=5e-6, wd=0.001)
        learn.fit(1, lr=5e-6, wd=0.001)


# + {"colab_type": "code", "id": "JTqfMd9JUQ1k", "colab": {}}
    FLAGS={}
    #xmp.spawn(train_loop, args=(FLAGS,),  nprocs=8, start_method='fork')


# + {"colab_type": "code", "id": "oh5655pAUQ10", "colab": {}}
    import pysnooper

# + {"colab_type": "code", "id": "2YGBLtQ1UQ12", "colab": {}}
    @pysnooper.snoop()
    def _mp_fn(rank, flags, k=k):
        device = xm.xla_device(devkind='TPU')
        logger.debug("%s used for xla_device" % device)
        net = k.model
        net.to(device)
        logger.debug("%s used for xla_device, to device done" % device)

        train_sampler = DistributedSamplerWrapper(
            sampler=BalanceClassSampler(labels=k.train_dataset.get_labels(), mode="downsampling"),
            num_replicas=xm.xrt_world_size(),
            rank=xm.get_ordinal(),
            shuffle=True
        )
        train_loader = torch.utils.data.DataLoader(
            k.train_dataset,
            batch_size=TrainGlobalConfig.batch_size,
            sampler=train_sampler,
            pin_memory=False,
            drop_last=True,
            num_workers=TrainGlobalConfig.num_workers,
        )
        validation_sampler = torch.utils.data.distributed.DistributedSampler(
            k.validation_dataset,
            num_replicas=xm.xrt_world_size(),
            rank=xm.get_ordinal(),
            shuffle=False
        )
        validation_loader = torch.utils.data.DataLoader(
            k.validation_dataset,
            batch_size=TrainGlobalConfig.batch_size,
            sampler=validation_sampler,
            pin_memory=False,
            drop_last=False,
            num_workers=TrainGlobalConfig.num_workers
        )
        validation_tune_sampler = torch.utils.data.distributed.DistributedSampler(
            k.validation_tune_dataset,
            num_replicas=xm.xrt_world_size(),
            rank=xm.get_ordinal(),
            shuffle=True
        )
        validation_tune_loader = torch.utils.data.DataLoader(
            k.validation_tune_dataset,
            batch_size=TrainGlobalConfig.batch_size,
            sampler=validation_tune_sampler,
            pin_memory=False,
            drop_last=False,
            num_workers=TrainGlobalConfig.num_workers
        )
        test_sampler = torch.utils.data.distributed.DistributedSampler(
            k.test_dataset,
            num_replicas=xm.xrt_world_size(),
            rank=xm.get_ordinal(),
            shuffle=False
        )
        test_loader = torch.utils.data.DataLoader(
            k.test_dataset,
            batch_size=TrainGlobalConfig.batch_size,
            sampler=test_sampler,
            pin_memory=False,
            drop_last=False,
            num_workers=TrainGlobalConfig.num_workers
        )

        logger.debug("rank: %d. Will create TPU Fitter", rank)

        if rank == 0:
            time.sleep(1)

        fitter = TPUFitter(model=net, device=device, config=TrainGlobalConfig)
        fitter.fit(train_loader, validation_loader)
        fitter.run_tuning_and_inference(test_loader, validation_tune_loader)




# + {"colab_type": "code", "id": "cVRPNFYuUQ18", "colab": {}}
    import gc
    gc.collect()


# + {"colab_type": "code", "id": "6x45fTI1UQ2B", "colab": {}}

    FLAGS={}
    xmp.spawn(_mp_fn, args=(FLAGS,),  nprocs=8, start_method='fork')

# + {"colab_type": "code", "id": "JJPxfj1OUQ2E", "colab": {}}
    from kaggle_runner.kernels.kernel import KaggleKernelOnlyPredict



# + {"colab_type": "code", "id": "a_etfRgnUQ2I", "colab": {}}
    from datetime import date
    today = date.today()
    output_model_file='XLMRobertaModel_tpu_trained.bin'
    torch.save(k.model.state_dict(), f"{today}_{output_model_file}")


# + {"colab_type": "code", "id": "6Bu-tn3lUQ2J", "colab": {}}
    submission = pd.concat([pd.read_csv(path) for path in glob('node_submissions/*.csv')]).groupby('id').mean()
    submission['toxic'].hist(bins=100)


# + {"colab_type": "code", "id": "1G896nEvUQ2L", "colab": {}}
    submission.to_csv(f'{ROOT_PATH}/submission.csv')


# + {"colab_type": "code", "id": "8RHAQ4PNUQ2N", "colab": {}}
# #!cp log.txt '/content/drive/My Drive/jigsaw2020-kaggle-public-baseline/'
# !make push_dataset



import math
import os
from glob import glob

import numpy as np
import pandas as pd
import pydicom
from matplotlib import pyplot as plt
from sklearn.model_selection import KFold
from tensorflow.keras.layers import (Conv2D, Conv2DTranspose, Dropout, Input,
                                     MaxPooling2D, concatenate)
from tensorflow.keras.models import Model
from tensorflow.python.ops import math_ops
from tqdm import tqdm

import kaggle_runner.datasets.data_handlers
from kaggle_runner.kernels.kernel import KaggleKernel
from kaggle_runner.post_processers import PS_result_analyzer
from kaggle_runner.utils import kernel_utils


class PS(KaggleKernel):
    def __init__(self, *args, **kargs):
        super(PS, self).__init__(*args, **kargs)
        self._PS_init_()

    def _PS_init_(self):
        self.use_tf_data = False
        self.developing = True
        # self.train_dev_mask = None  # for train dev split
        self.DATA_PATH_BASE = "../input/siim-train-test"
        self._im_chan = 1

        self.BS = 16
        try:
            self.ds  # if has this attribute, no need to overwrite
        except AttributeError:
            self.ds = None  # for train
            self.ds_len = 0
            self.dev_ds = None
            self.dev_ds_len = 0
            self.tf_data_handler = None

    def analyze_data(self):
        pass

    @staticmethod
    def _PS_data_preprocess_np(fns, df, TARGET_COLUMN, im_height, im_width, im_chan):
        X_train = np.zeros((len(fns), im_height, im_width, im_chan), dtype=np.uint8)
        Y_train = np.zeros((len(fns), im_height, im_width, 1), dtype=np.uint8)
        print("Getting train images and masks ... ")
        # sys.stdout.flush()

        for n, _id in tqdm(enumerate(fns), total=len(fns)):
            dataset = pydicom.read_file(_id)
            _id_keystr = _id.split("/")[-1][:-4]
            X_train[n] = np.expand_dims(dataset.pixel_array, axis=2)
            try:
                mask_data = df.loc[_id_keystr, TARGET_COLUMN]

                if "-1" in mask_data:
                    Y_train[n] = np.zeros((1024, 1024, 1))
                else:
                    if type(mask_data) == str:
                        Y_train[n] = np.expand_dims(
                            kernel_utils.rle2mask(
                                df.loc[_id_keystr, TARGET_COLUMN], 1024, 1024
                            ).T,
                            axis=2,
                        )
                    else:
                        Y_train[n] = np.zeros((1024, 1024, 1))

                        for x in mask_data:
                            Y_train[n] = Y_train[n] + np.expand_dims(
                                kernel_utils.rle2mask(x, 1024, 1024).T, axis=2
                            )
            except KeyError:
                print(
                    f"Key {_id.split('/')[-1][:-4]} without mask, assuming healthy patient."
                )
                # Assume missing masks are empty masks.
                Y_train[n] = np.zeros((1024, 1024, 1))

        print("Done data preprocessing as numpy array!")

        return X_train, Y_train

    @staticmethod
    def _PS_data_preprocess(fns, df, tf=False):
        """
        for tf=True, need to use TF2.0

        :param fns:
        :param df:
        :param tf:
        :return:
        """
        TARGET_COLUMN = " EncodedPixels"
        im_height = 1024
        im_width = 1024
        im_chan = 1
        # Get train images and masks

        if not tf:
            return PS._PS_data_preprocess_np(
                fns, df, TARGET_COLUMN, im_height, im_width, im_chan
            )
        else:
            return PS._PS_data_preprocess_tf(
                fns, df, TARGET_COLUMN, im_height, im_width, im_chan
            )

    @staticmethod
    def _PS_data_preprocess_tf(fns, df, TARGET_COLUMN, im_height, im_width, im_chan):
        tf_data_handler = kaggle_runner.datasets.data_handlers.PS_TF_DataHandler()

        return tf_data_handler.to_tf_from_disk(
            fns, df, TARGET_COLUMN, im_height, im_width, im_chan
        )

    @staticmethod
    def img_build_patches(imgs):
        im_height = 128
        im_width = 128
        imgs = imgs.reshape((-1, im_height, im_width, 1))

        return imgs

    def prepare_train_dev_data(self):
        if self.use_tf_data:
            self._prepare_train_data_as_tf()
        else:
            self._prepare_train_data_as_np()

    def _get_train_data(self):
        train_data_wildcard = self.DATA_PATH_BASE + "/dicom-images-train/*/*/*.dcm"

        if self.developing:
            train_fns = sorted(glob(train_data_wildcard))[:100]
        else:
            train_fns = sorted(glob(train_data_wildcard))

        return train_fns

    # just prepare train/dev here together
    def _prepare_train_data_as_tf(self):
        train_fns = self._get_train_data()

        self.logger.debug(f"train & dev counts: {len(train_fns)}")
        df_full = pd.read_csv(
            os.path.join(self.DATA_PATH_BASE, "train-rle.csv"), index_col="ImageId"
        )
        self.ds = PS._PS_data_preprocess(train_fns, df_full, tf=True)

    # just prepare train/dev here together
    def _prepare_train_data_as_np(self):
        # getting path of all the train and test images

        train_fns = self._get_train_data()

        self.logger.debug(f"train & dev counts: {len(train_fns)}")
        df_full = pd.read_csv(
            os.path.join(self.DATA_PATH_BASE, "train-rle.csv"), index_col="ImageId"
        )

        train_fns_splits = []
        idx = 0
        LEN = 1024

        for idx in range(len(train_fns) // LEN + 1):
            train_fns_splits.append(train_fns[idx * LEN : idx * LEN + LEN])

        for i, fns in enumerate(train_fns_splits):
            images, mask_e = PS._PS_data_preprocess(fns, df_full)

            # just use sklearn split to get id and it is fine. For text thing,
            splits = list(
                KFold(n_splits=5, random_state=2019, shuffle=True).split(images)
            )
            # just do 1 fold, later we can add them all back
            tr_ind, val_ind = splits[0]

            self.train_X = images[tr_ind]
            self.train_Y = mask_e[tr_ind]

            self.dev_X = images[val_ind]
            self.dev_Y = mask_e[val_ind]

            self.logger.debug(
                f"size: train_X {self.train_X.shape}, train_Y {self.train_Y.shape}"
            )

            self.logger.debug(sorted(glob("*pkl")))  # list file now
            # self.save_data_tf(file_name=f"train_dev.{i}.tfrec")
            # del self.train_X # not del if not save to tfrec
            # del self.train_Y
            # del self.dev_X
            # del self.dev_Y
            # del images
            # del mask_e
            # gc.collect()
            # self.logger.debug(sorted(glob("*")))

    def _clone_data(self, src):
        self.model = src.model
        self.model_metrics = src.model_metrics
        self.model_loss = src.model_loss

        self.train_X = src.train_X
        self.train_Y = src.train_Y

        self.dev_X = src.dev_X
        self.dev_Y = src.dev_Y

        self.test_X = src.test_X

        self.result_analyzer = src.result_analyzer  # for analyze the result

        self._stage = src._stage

    def save_data_tf(self, file_name="train_dev.tfrec"):
        assert self.train_X is not None

        all_X = np.concatenate((self.train_X, self.dev_X), axis=0)
        all_Y = np.concatenate((self.train_Y, self.dev_Y), axis=0)

        ds = kaggle_runner.datasets.data_handlers.PS_TF_DataHandler.get_train_dataset(
            all_X, all_Y
        )

        # def train_input_fn_bt(features, labels, batch_size, cv, split_id=None, n_splits=None, ds=None):
        assert len(all_X) == len(all_Y)
        try:
            self.BS
        except AttributeError:
            self._PS_init_()
        split = 5
        # self.ds = utils.PS_TF_DataHandler.train_input_fn_bt(None, None, BS, cv=True, split_id=0, n_splits=split, ds=ds, ds_len=math.floor(len(all_X)*(1-1/split)))
        self.ds = ds
        self.ds_len = math.floor(len(all_X) * (1 - 1 / split))
        # self.dev_ds = utils.PS_TF_DataHandler.train_input_fn_bt(None, None, BS, cv=True, cv_train=False, split_id=0, n_splits=split, ds=ds, ds_len=math.floor(len(all_X) * (1 / split)))
        self.dev_ds_len = math.floor(len(all_X) * (1 / split))

        self.logger.debug(f"dev set counts: {self.dev_ds_len} / {len(all_X)}")

        kaggle_runner.datasets.data_handlers.PS_TF_DataHandler.to_tfrecord(
            self.ds, file_name=file_name
        )

    def after_prepare_data_hook(self):
        self.analyze_data()

    def train_model(self):
        # self.model.fit(self.train_X, self.train_Y, validation_data=(self.dev_X, self.dev_Y), batch_size=256, epochs=5)  # can run now...
        v_s = math.floor(self.dev_ds_len / self.BS)

        def split_train_dev(ds, n_splits, split_id):
            ds_shards = [ds.shard(n_splits, i) for i in range(n_splits)]

            shards_cross = [
                ds_shards[val_id] for val_id in range(n_splits) if val_id != split_id
            ]

            ds_train = shards_cross[0]

            for t in shards_cross[1:]:
                ds_train = ds.concatenate(t)

            return ds, ds.shard(n_splits, split_id)

        def mask_to_binary(a, b, threshold=0.5):
            # b = tf.transpose(b, [1,0,2])
            threshold = math_ops.cast(threshold, b.dtype)
            b = math_ops.cast(b > threshold, b.dtype)

            return a, b

        # TODO ref put ds prepare to right partition
        def ds_prepare():
            ds = self.ds
            ds = ds.map(mask_to_binary)
            ds, ds_dev = split_train_dev(ds, 5, 0)
            ds = ds.shuffle(buffer_size=1024).repeat().batch(self.BS).prefetch(10)
            ds_dev = ds_dev.batch(self.BS).prefetch(10)
            self.ds = ds
            self.dev_ds = ds_dev

        ds_prepare()
        self.model.fit(
            self.ds,
            steps_per_epoch=1024,
            epochs=1,
            validation_data=self.dev_ds,
            validation_steps=v_s if v_s > 0 else 1,
            verbose=1,
        )

    def prepare_test_data(self):
        test_data_wildcard = self.DATA_PATH_BASE + "/dicom-images-test/*/*/*.dcm"

        if self.developing:
            test_fns = sorted(glob(test_data_wildcard))[:100]
        else:
            test_fns = sorted(glob(test_data_wildcard))
        self.logger.debug(f"test counts: {len(test_fns)}")

    def _build_model(self, input_layer, start_neurons):
        # ref: https://www.kaggle.com/phoenigs/u-net-dropout-augmentation-stratification
        # 128 -> 64
        conv1 = Conv2D(start_neurons * 1, (3, 3), activation="relu", padding="same")(
            input_layer
        )
        conv1 = Conv2D(start_neurons * 1, (3, 3), activation="relu", padding="same")(
            conv1
        )
        pool1 = MaxPooling2D((2, 2))(conv1)
        pool1 = Dropout(0.25)(pool1)

        # 64 -> 32
        conv2 = Conv2D(start_neurons * 2, (3, 3), activation="relu", padding="same")(
            pool1
        )
        conv2 = Conv2D(start_neurons * 2, (3, 3), activation="relu", padding="same")(
            conv2
        )
        pool2 = MaxPooling2D((2, 2))(conv2)
        pool2 = Dropout(0.5)(pool2)

        # 32 -> 16
        conv3 = Conv2D(start_neurons * 4, (3, 3), activation="relu", padding="same")(
            pool2
        )
        conv3 = Conv2D(start_neurons * 4, (3, 3), activation="relu", padding="same")(
            conv3
        )
        pool3 = MaxPooling2D((2, 2))(conv3)
        pool3 = Dropout(0.5)(pool3)

        # 16 -> 8
        conv4 = Conv2D(start_neurons * 8, (3, 3), activation="relu", padding="same")(
            pool3
        )
        conv4 = Conv2D(start_neurons * 8, (3, 3), activation="relu", padding="same")(
            conv4
        )
        pool4 = MaxPooling2D((2, 2))(conv4)
        pool4 = Dropout(0.5)(pool4)

        # Middle
        convm = Conv2D(start_neurons * 16, (3, 3), activation="relu", padding="same")(
            pool4
        )
        convm = Conv2D(start_neurons * 16, (3, 3), activation="relu", padding="same")(
            convm
        )

        # 8 -> 16
        deconv4 = Conv2DTranspose(
            start_neurons * 8, (3, 3), strides=(2, 2), padding="same"
        )(convm)
        uconv4 = concatenate([deconv4, conv4])
        uconv4 = Dropout(0.5)(uconv4)
        uconv4 = Conv2D(start_neurons * 8, (3, 3), activation="relu", padding="same")(
            uconv4
        )
        uconv4 = Conv2D(start_neurons * 8, (3, 3), activation="relu", padding="same")(
            uconv4
        )

        # 16 -> 32
        deconv3 = Conv2DTranspose(
            start_neurons * 4, (3, 3), strides=(2, 2), padding="same"
        )(uconv4)
        uconv3 = concatenate([deconv3, conv3])
        uconv3 = Dropout(0.5)(uconv3)
        uconv3 = Conv2D(start_neurons * 4, (3, 3), activation="relu", padding="same")(
            uconv3
        )
        uconv3 = Conv2D(start_neurons * 4, (3, 3), activation="relu", padding="same")(
            uconv3
        )

        # 32 -> 64
        deconv2 = Conv2DTranspose(
            start_neurons * 2, (3, 3), strides=(2, 2), padding="same"
        )(uconv3)
        uconv2 = concatenate([deconv2, conv2])
        uconv2 = Dropout(0.5)(uconv2)
        uconv2 = Conv2D(start_neurons * 2, (3, 3), activation="relu", padding="same")(
            uconv2
        )
        uconv2 = Conv2D(start_neurons * 2, (3, 3), activation="relu", padding="same")(
            uconv2
        )

        # 64 -> 128
        deconv1 = Conv2DTranspose(
            start_neurons * 1, (3, 3), strides=(2, 2), padding="same"
        )(uconv2)
        uconv1 = concatenate([deconv1, conv1])
        uconv1 = Dropout(0.5)(uconv1)
        uconv1 = Conv2D(start_neurons * 1, (3, 3), activation="relu", padding="same")(
            uconv1
        )
        uconv1 = Conv2D(start_neurons * 1, (3, 3), activation="relu", padding="same")(
            uconv1
        )

        # uconv1 = Dropout(0.5)(uconv1)
        output_layer = Conv2D(1, (1, 1), padding="same", activation="sigmoid")(uconv1)

        return output_layer

    def build_and_set_model(self):

        img_size = 256
        input_layer = Input((img_size, img_size, 1))
        output_layer = self._build_model(input_layer, 16)
        model = Model(input_layer, output_layer)
        model.compile(
            loss="binary_crossentropy",
            optimizer="adam",
            metrics=["accuracy", kernel_utils.dice_coef],
        )

        self.model = model

    def set_result_analyzer(self):
        self.result_analyzer = PS_result_analyzer()

    @staticmethod
    def _check_image_data(ds):
        cnt = 0

        for image, mask in ds:
            print(cnt)
            cnt += 1
            m = mask.numpy()

            if m.max() <= 0:
                continue
            m = np.reshape(m, (1024, 1024))
            img = image.numpy()
            img = np.reshape(img, (1024, 1024))
            # plt.imshow(img)
            plt.imshow(m)

            break

    def _recover_from_tf(self):
        self.ds = kaggle_runner.datasets.data_handlers.PS_TF_DataHandler.from_tfrecord()
        PS._check_image_data(self.ds)



from functools import partial, wraps

from kaggle_runner import may_debug


def size_decorator(f):
    @wraps(f)
    def size_wrapper(y_true, y_pred, *args, **kwargs):
        ts = y_true.shape
        ps = y_pred.shape

        if len(ps) > len(ts) or (
                (ps[1] > ts[1]) if ts[1] is not None else False):
            y_pred = y_pred[:,0]

        return f(y_true, y_pred, *args, **kwargs)

    return size_wrapper



from googletrans import Translator

import plotly.graph_objects as go
from colorama import Fore, Style
from kaggle_runner import may_debug

# ### Visualize model predictions
#
# Now, I will visualize the performance of the model on few validation samples.

# + {"_kg_hide-input": true}
translator = Translator()

def visualize_model_preds(model,val_data, x_valid, y_valid, indices=[0, 17, 1, 24]):
    comments = val_data.comment_text.loc[indices].values.tolist()
    preds = model.predict(x_valid[indices].reshape(len(indices), -1))[:,0]

    for idx, i in enumerate(indices):
        if y_valid[i] == 0:
            label = "Non-toxic"
            color = f'{Fore.GREEN}'
        else:
            label = "Toxic"
            color = f'{Fore.RED}'

        print('{}{}'.format(color, str(idx+1) + ". " + label))
        print(f'{Style.RESET_ALL}')

        try:
            print(f"Predicted {preds[idx]}")

            print("ORIGINAL")
            print(comments[idx].encode('utf-8')); print("")
            print("TRANSLATED")
            print(translator.translate(comments[idx]).text)
        except Exception as e:
            print(e)
            print("encode error")

        # fig = go.Figure()

        # if list.index(sorted(preds[:, 0]), preds[idx][0]) > 1:
        #     yl = [preds[idx][0], 1 - preds[idx][0]]
        # else:
        #     yl = [1 - preds[idx][0], preds[idx][0]]
        # fig.add_trace(go.Bar(x=['Non-Toxic', 'Toxic'], y=yl,
        #     marker=dict(color=["seagreen", "indianred"])))
        # fig.update_traces(name=comments[idx])
        # fig.update_layout(xaxis_title="Labels", yaxis_title="Probability",
        #     template="plotly_white", title_text=("Predictions for validation "
        #     "comment #{}").format(idx+1))
        # fig.show()



import tensorflow.keras.backend as K
from tensorflow.keras import constraints, initializers, regularizers
from tensorflow.keras.layers import (
    Activation,
    BatchNormalization,
    Bidirectional,
    Dense,
    Embedding,
    GlobalAveragePooling1D,
    GlobalMaxPooling1D,
    Input,
    Lambda,
    Layer,
    PReLU,
    SpatialDropout1D,
    add,
    concatenate,
)

from kaggle_runner.kernels.KernelRunningState import KernelRunningState
from kaggle_runner.utils.kernel_utils import dump_obj, get_obj_or_dump, logger


class AttentionRaffel(Layer):
    def __init__(
        self,
        step_dim,
        W_regularizer=None,
        b_regularizer=None,
        W_constraint=None,
        b_constraint=None,
        bias=True,
        **kwargs,
    ):
        """
        Follows the work of Raffel et al. [https://arxiv.org/abs/1512.08756]
        :param step_dim: feature vector length
        :param W_regularizer:
        :param b_regularizer:
        :param W_constraint:
        :param b_constraint:
        :param bias:
        :param kwargs:
        """
        super(AttentionRaffel, self).__init__(**kwargs)
        self.supports_masking = True
        self.init = "glorot_uniform"

        self.W_regularizer = regularizers.get(W_regularizer)
        self.b_regularizer = regularizers.get(b_regularizer)

        self.W_constraint = constraints.get(W_constraint)
        self.b_constraint = constraints.get(b_constraint)

        self.bias = bias
        self.step_dim = step_dim
        self.features_dim = 0

    def get_config(self):
        config = {
            "step_dim": self.step_dim,
            "bias": self.bias,
            "W_regularizer": regularizers.serialize(self.W_regularizer),
            "b_regularizer": regularizers.serialize(self.b_regularizer),
            "W_constraint": constraints.serialize(self.W_constraint),
            "b_constraint": constraints.serialize(self.b_constraint),
        }
        base_config = super(AttentionRaffel, self).get_config()
        if "cell" in base_config:
            del base_config["cell"]
        return dict(list(base_config.items()) + list(config.items()))

    def build(self, input_shape):
        # Input shape 3D tensor with shape: `(samples, steps, features)`.
        # one step is means one bidirection?
        assert len(input_shape) == 3

        self.W = self.add_weight(
            "{}_W".format(self.name),
            (int(input_shape[-1]),),
            initializer=self.init,
            regularizer=self.W_regularizer,
            constraint=self.W_constraint,
        )
        self.features_dim = input_shape[-1]  # features dimention of input

        if self.bias:
            self.b = self.add_weight(
                "{}_b".format(self.name),
                (int(input_shape[1]),),
                initializer="zero",
                regularizer=self.b_regularizer,
                constraint=self.b_constraint,
            )
        else:
            self.b = None

        self.built = True

    def compute_mask(self, inputs, mask=None):
        # do not pass the mask to the next layers
        return None

    def call(self, x, mask=None):
        # more like the alignment model, which scores how the inputs around
        # position j and the output at position i match
        features_dim = self.features_dim
        step_dim = self.step_dim

        eij = K.reshape(
            K.dot(
                K.reshape(x, (-1, features_dim)
                          ), K.reshape(self.W, (features_dim, 1))
            ),
            (-1, step_dim),
        )

        if self.bias:
            eij += self.b

        eij = K.tanh(eij)  # activation

        # softmax
        a = K.exp(eij)

        # apply mask after the exp. will be re-normalized next
        if mask is not None:
            # Cast the mask to floatX to avoid float64 upcasting in theano
            a *= K.cast(mask, K.floatx())

        # in some cases especially in the early stages of training the sum may
        # be almost zero
        a /= K.cast(K.sum(a, axis=1, keepdims=True) + K.epsilon(), K.floatx())

        a = K.expand_dims(a)
        # context vector c_i (or for this, only one c_i)
        weighted_input = x * a
        # print weigthted_input.shape
        return K.sum(weighted_input, axis=1)

    def compute_output_shape(self, input_shape):
        # return input_shape[0], input_shape[-1]
        return input_shape[0], self.features_dim

    def set_metrics(self):
        """
        set_metrics for model training

        :return: None
        """
        pass

    def set_result_analyzer(self):
        pass

    def pre_prepare_data_hook(self):
        pass

    def after_prepare_data_hook(self):
        pass

    def prepare_train_dev_data(self):
        pass

    def prepare_test_data(self):
        pass

    def predict_on_test(self):
        pass

    def dump_state(self, exec_flag=False, force=True):
        logger.debug(f"state {self._stage}")
        if exec_flag:
            logger.debug(f"dumping state {self._stage}")
            dump_obj(self, f"run_state_{self._stage}.pkl", force=force)
            # dump_obj(self, 'run_state.pkl', force=True)  # too large

    def run(
        self,
        start_stage=None,
        end_stage=KernelRunningState.SAVE_SUBMISSION_DONE,
        dump_flag=False,
        force_dump=True,
    ):
        """

        :param start_stage: if set, will overwrite the stage
        :param end_stage:
        :param dump_flag:
        :return:
        """
        self.continue_run(
            start_stage=start_stage,
            end_stage=end_stage,
            dump_flag=dump_flag,
            force_dump=force_dump,
        )

    def continue_run(
        self,
        start_stage=None,
        end_stage=KernelRunningState.SAVE_SUBMISSION_DONE,
        dump_flag=False,
        force_dump=True,
    ):
        if start_stage is not None:
            assert start_stage.value < end_stage.value
            self._stage = start_stage

        if self._stage.value < KernelRunningState.PREPARE_DATA_DONE.value:
            self.pre_prepare_data_hook()
            self.prepare_train_dev_data()
            self.after_prepare_data_hook()

            self._stage = KernelRunningState.PREPARE_DATA_DONE
            self.dump_state(exec_flag=dump_flag, force=force_dump)
            if self._stage.value >= end_stage.value:
                return

        if self._stage.value < KernelRunningState.TRAINING_DONE.value:
            self.pre_train()
            self.build_and_set_model()
            self.train_model()
            self.after_train()

            self.save_model()

            self._stage = KernelRunningState.TRAINING_DONE
            self.dump_state(exec_flag=dump_flag, force=force_dump)
            if self._stage.value >= end_stage.value:
                return

        if self._stage.value < KernelRunningState.EVL_DEV_DONE.value:
            self.set_result_analyzer()

            self._stage = KernelRunningState.EVL_DEV_DONE
            self.dump_state(exec_flag=False, force=force_dump)
            if self._stage.value >= end_stage.value:
                return

        if self._stage.value < KernelRunningState.SAVE_SUBMISSION_DONE.value:
            self.pre_test()
            self.prepare_test_data()
            self.predict_on_test()
            self.after_test()

            self._stage = KernelRunningState.SAVE_SUBMISSION_DONE
            self.dump_state(exec_flag=False, force=force_dump)
            if self._stage.value >= end_stage.value:
                return

    @classmethod
    def _load_state(cls, stage=None, file_name="run_state.pkl"):
        """

        :param file_name:
        :return: the kernel object, need to continue
        """
        if stage is not None:
            file_name = f"run_state_{stage}.pkl"
        logger.debug(f"restore from {file_name}")
        return get_obj_or_dump(filename=file_name)

    def load_state_data_only(self, file_name="run_state.pkl"):
        pass

    @classmethod
    def load_state_continue_run(cls, file_name="run_state.pkl"):
        """

        :param file_name:
        :return: the kernel object, need to continue
        """
        self = cls._load_state(file_name=file_name)
        self.continue_run()

    def pre_train(self):
        pass

    def after_train(self):
        pass

    def pre_test(self):
        pass

    def after_test(self):
        pass



from sklearn.metrics import accuracy_score, roc_auc_score
from tensorflow.keras.callbacks import (Callback, CSVLogger, ModelCheckpoint,
                                        ReduceLROnPlateau)

from kaggle_runner import logger
from kaggle_runner.metrics.meters import AverageMeter,RocAucMeter


class RocAucEvaluation(Callback):
    def __init__(self, validation_data=(), interval=1):
        super(Callback, self).__init__()

        self.interval = interval
        self.X_val, self.y_val = validation_data

    def on_epoch_end(self, epoch, logs={}):
        if epoch % self.interval == 0:
            y_pred = self.model.predict(self.X_val, verbose=0)

            if y_pred.size > self.y_val.size:
                y_pred = y_pred[:,0]
            score = roc_auc_score(self.y_val, y_pred)
            print(
                "\n ROC-AUC - epoch: {:d} - score: {:.6f}".format(epoch+1, score))
            logger.debug(
                "\n ROC-AUC - epoch: {:d} - score: {:.6f}".format(epoch+1, score))


def ReduceLROnPlateauLogCBs(validation_data):
    cb = []

    reduceLROnPlat = ReduceLROnPlateau(monitor='val_loss',
                                       factor=0.3, patience=3,
                                       verbose=1, mode='auto',
                                       epsilon=0.0001, cooldown=1, min_lr=0.000001)
    cb.append(reduceLROnPlat)
    log = CSVLogger('log.csv')
    cb.append(log)

    RocAuc = RocAucEvaluation(validation_data=validation_data, interval=1)
    cb.append(RocAuc)
    # ckpt = ModelCheckpoint("toxic.hdf5", save_best_only=True, verbose=1) bert get_config not implemented, cannot save
    # cb.append(ckpt)

    return cb

import torch
from fastai.core import Any
from fastai.basic_train import LearnerCallback, Learner

def _check_grad(raw_opt):
    pg = raw_opt.param_groups
    pg0pl = pg[0]['params'] # pg0pl[0] is a Parameter
    pg1pl = pg[1]['params'] # pg0pl[0] is a Parameter

    with torch.no_grad():
        #norms = torch.tensor([torch.norm(p) for p in pg0pl])
        #may_debug()
        #logger.debug("%s", pg0pl[0].grad)
        #logger.debug("%s", pg0pl[0].data)
        normsg = torch.tensor([torch.norm(p.grad) for p in pg0pl[:10] if p.grad is not None])
        #logger.debug("params info pg0: norm std(%f) mean(%f)", *torch.std_mean(norms))
        logger.debug("grad info pg0: norm std(%f) mean(%f)", *torch.std_mean(normsg))

        #norms1 = torch.tensor([torch.norm(p) for p in pg1pl])
        norms1g = torch.tensor([torch.norm(p.grad) for p in pg1pl[:10] if p.grad is not None])
        #logger.debug("params info pg1: norm std(%f) mean(%f)", *torch.std_mean(norms1))
        logger.debug("grad info pg1: norm std(%f) mean(%f)", *torch.std_mean(norms1g))

class CheckGrad(LearnerCallback):
    def __init__(self, learn:Learner, skip_loss_step=False, batch_size=16):
        super().__init__(learn)
        self.skip_loss_step = skip_loss_step
        logger.debug("Init Callback CheckGrad with skip_loss_step: " +str(self.skip_loss_step))
        self.losses = None
        self.final_scores = None
        self.batch_size = batch_size

    def on_train_begin(self, **kwargs:Any)->None:
        self.losses = AverageMeter()
        self.final_scores = RocAucMeter()

    def on_backward_begin(self, **kwargs:Any)->None:
        #print(kwargs.keys())
        """dict_keys(['epoch', 'iteration', 'num_batch', 'skip_validate',
        'n_epochs', 'pbar', 'metrics', 'stop_training', 'last_input',
        'last_target', 'train', 'stop_epoch', 'skip_step', 'skip_zero',
        'skip_bwd', 'last_output', 'last_loss', 'smooth_loss'])
        """
        pg = self.learn.opt.opt.param_groups
        #logger.debug("grad info: %s", raw_opt)
        logger.debug(f"on_backward_begin lr: {pg[0]['lr']}")
        logger.debug("itr: %d, num_batch: %d, last loss: %f, smooth_loss: %f",
                     kwargs['iteration'], kwargs['num_batch'],
                     kwargs['last_loss'], kwargs['smooth_loss'])

        self.final_scores.update(kwargs['last_target'], kwargs['last_output'])
        self.losses.update(kwargs['last_loss'].detach().item(), self.batch_size)
        logger.debug(f"loss_avg: {self.losses.avg:.5f}, lr_pg0:"
                     f"{pg[0]['lr']}, lr_pg1: {pg[1]['lr']}final_score:"
                     f"{self.final_scores.avg:.5f}, mc_score:"
                     f"{self.final_scores.mc_avg:.5f}")

    def on_backward_end(self, **kwargs:Any)->None:
        raw_opt = self.learn.opt.opt
        _check_grad(raw_opt)

        return {'skip_step': self.skip_loss_step}



import os

import cv2
from torch.utils.data import DataLoader, Dataset, sampler

from albumentations import (Blur, Compose, ElasticTransform, GaussNoise,
                            GridDistortion, HorizontalFlip, IAAEmboss,
                            MultiplicativeNoise, Normalize, OneOf,
                            OpticalDistortion, RandomBrightnessContrast,
                            RandomGamma, RandomRotate90, Resize,
                            ShiftScaleRotate, Transpose, VerticalFlip)
from albumentations.pytorch import ToTensor


class MockDataset(Dataset):
    def __init__(self, root, df, size, mean, std, tta=4):
        self.root = root
        self.size = size
        self.fnames = list(df["ImageId"])
        self.num_samples = len(self.fnames)
        self.transform = Compose(
            [Normalize(mean=mean, std=std, p=1), Resize(size, size), ToTensor(),]
        )

    def __getitem__(self, idx):
        fname = self.fnames[idx]
        path = os.path.join(self.root, fname + ".png")
        image = cv2.imread(path)
        images = self.transform(image=image)["image"]
        return images

    def __len__(self):
        return self.num_samples
